/*****************************************************************
*
* hal header
*
* Copyright (c) 2013 Dr. Neuhaus Telekommunikation GmbH
*
*
*****************************************************************/

#ifndef __HAL_H__
#define __HAL_H__

#include <stdio.h>
#include <assert.h>
#include <stdint.h>

#include <string>
#include <map>
#include <list>

#define ON		true
#define OFF		false
#define UP		true
#define DOWN	false

#define PWR_LED_GREEN 0
#define PWR_LED_BLUE  3
#define WAN_LED_GREEN 1
#define WAN_LED_BLUE  4
#define LMN_LED_GREEN 2
#define LMN_LED_BLUE  5

namespace hal
{
	bool halInit();
	void halDeinit();
	
	/// <unfinished />
	/// 
	/// <summary>
	/// Schalten der LEDs
	/// </summary>
	/// <sfr>
	/// <b>enforced supported not interfered</b>
	/// </sfr>
	/// <value>
	/// Ein-/Ausschalten der led0 bis led5
	/// </value>
	/// <param name="_led">Nr. der LED</param>
	/// <param name="_on">ON (true) fr Einschalten, OFF (false) fr Ausschalten</param>
	/// <returns>
	/// true bei erfolgreicher Ausfhrung, false bei aufgetretenem Fehler
	/// </returns>		
	bool SwitchLed(int _led, bool _on);

	/// <summary>
	/// Watchdog des Systems (Hardware) triggern
	/// </summary>

	/// <unfinished />
	/// 
	/// <summary>
	/// Triggern des Watchdogs
	/// </summary>
	/// <sfr>
	/// <b>enforced supported not interfered</b>
	/// </sfr>
	/// <value>
	/// Triggert den/die Watchdogs. 
	/// Bei Vorhandensein von /dev/watchdog1 (Softdog) werden dieser und /dev/watchdog0 (Harddog) 
	/// getriggert (Heartbeat Softdog: 20s, Heartbeat Harddog: 30s).
	/// Lsst sich /dev/watchdog1 nicht ffnen, wird nur der Harddog (/dev/watchdog) getriggert).
	/// (Heartbeat 19s)
	/// </value>
	/// <param name="_paramName">Parameter Beschreibung</param>
	/// <returns>
	/// Fehlerfall -1, Gutfall 0
	/// </returns>	
	/// <remarks>
	/// Anmerkungen oder lschen
	/// </remarks>
	int TriggerSysWD();
	
	/// <unfinished />
	/// 
	/// <summary>
	/// Abfrage des Systemspeichers
	/// </summary>
	/// <sfr>
	/// <b>enforced supported not interfered</b>
	/// </sfr>
	/// <value>
	/// Ermittelt die Gre des RAM-Speichers, die Gre des durch den Prozess beanspruchten RAM-Speicher
	/// und die Gre des noch zur Verfgung stehenden RAM Speichers
	/// </value>
	/// <param name="_total_KB">Gesamtgre des RAM-Speichers in KB</param>
	/// <param name="_avail_KB">Noch verfgbarer Speicher in KB</param>
	/// <param name="_process_KB">im RAM befindlicher Teil des Programmes - RSS - Resident Set Size</param>
	/// <returns>
	/// true bei erfolgreicher Ausfhrung, false bei aufgetretenem Fehler
	/// </returns>	
	bool GetSysMemory(uint64_t& _total_KB, uint64_t& _avail_KB, uint64_t& _process_KB);

	/// <summary>
	/// LED blinken lassen
	/// </summary>
	bool BlinkLed(int _led);

	/// <summary>
	/// Setzen des Blinkverhaltens der LED
	/// Paramter der Off-/Ontime knnen erst nach Auslsen des Blinkmodus gesetzt werden
	/// Das Beenden des Blinkmodus (led=off) setzt beide Parameter auf 500 ms zurck
	/// </summary>
	bool setLedDelay(int _led, const std::string& _delayofftime_ms, const std::string& _delayontime_ms);

	/// <summary>
	/// Setzen der IP-Adresse eines Eth-Devices
	/// Name des Devices in Kleinbuchstaben, z.Bsp: "eth0", "eth1.1"
	/// IPv4 als allgemein lesbaren string, z.Bsp: "192.168.168.168"
	/// </summary>
	int ConfigureIpAddr(const std::string& _ethname, const std::string& _ipv4addr);	

	/// <summary>
	/// Setzen der MAC-Adresse eines Eth-Devices
	/// Name des Devices in Kleinbuchstaben, z.Bsp: "eth0", "eth1.1"
	/// MAC als HEX, GROBUCHSTABEN, 6 Bytes, durch ':' getrennt, z.Bsp: "00:50:56:3f:3f:3f"
	/// Virtuelle netdevices erben die MAC vom physikalischem device
	/// anderweitige Konfiguration ist mglich, versetzt die phys. eth aber im Sniffermode
	/// </summary>
	int ConfigureIpHw(const std::string& _ethname, const std::string& _macaddr);

	/// <summary>
	/// Setzen der Netzmaske eines Eth-Devices
	/// Name des Devices in Kleinbuchstaben, z.Bsp: "eth0", "eth1.1"
	/// Netzmaske als allgemein lesbaren string in IPv4, z.Bsp: "255.255.0.0"
	/// </summary>
	int ConfigureIpNetmask(const std::string& _ethname, const std::string& _netmask);

	/// <summary>
	/// Setzen der Broadcast IP-Adresse eines Eth-Devices
	/// Name des Devices in Kleinbuchstaben, z.Bsp: "eth0", "eth1.1"
	/// IPv4 Broadcastadresse als allgemein lesbaren string, z.Bsp: "192.168.255.255"
	/// </summary>
	int ConfigureIpBroadcast(const std::string& _ethname, const std::string& _broadcast);
		
	/// <unfinished />
	/// 
	/// <summary>
	/// Setzt default gateway
	/// </summary>
	/// <sfr>
	/// <b>enforced supported not interfered</b>
	/// </sfr>
	/// <value>
	/// Setzen der Adresse des Default-Gateways, Zuordnung zum eth-Device erfolgt automatisch durch OS
	/// equivalent zu: route add default gw %_gateway%
	/// Das Setzen eines default gateways erfolgt nur bei entsprechend konfigurierter und aktiver Schnittstelle
	/// </value>
	/// <param name="_gateway">Gateway ipv4-Addr</param>
	/// <param name="_errortext">Beschreibung im Fehlerfall</param>
	/// <returns>
	/// true fr erfolgreich, false im Fehlerfall
	/// </returns>	
	bool ConfigureIpDefaultGateway(const std::string& _gateway, std::string& _errortext);

	/// <unfinished />
	/// 
	/// <summary>
	/// Entfernt ein default gateway
	/// </summary>
	/// <sfr>
	/// <b>enforced supported not interfered</b>
	/// </sfr>
	/// <value>
	/// Lscht ein default-Gateway mit der entsprechenden gatway-Addr
	/// equivalent zu: route del gw %_gateway%	
	/// </value>
	/// <param name="_paramName">Parameter Beschreibung</param>
	/// <returns>
	/// true fr erfolgreich, false im Fehlerfall
	/// </returns>		
	bool RemoveIPDefaultGateway(const std::string& _gateway, std::string& _errortext);
	
	/// <summary>
	/// Aktivieren/Deaktivieren eines netdevices
	/// Name des Devices in Kleinbuchstaben, z.Bsp: "eth0", "eth1.1"
	/// "true" aktiviert, "false" deaktiviert das netdevice
	/// </summary>
	int ConfigureIpUpDown(const std::string& _ethname, bool _ipupdown);

	/// <summary>
	/// Abfrage, ob auf Schicht des physical layers eine Verbindung zum eth-device besteht
	/// Name des Devices in Kleinbuchstaben, z.Bsp: "eth0", "eth1.1"
	/// "true" Link ist UP, "false" Link ist DOWN
	/// </summary>
	bool GetLinkState(const std::string& _ethname);

	/// <summary>
	/// Abfrage, ob das eth-device aus Sicht des OS aktiv oder inaktiv ist
	/// Name des Devices in Kleinbuchstaben, z.Bsp: "eth0", "eth1.1"
	/// "true" device ist UP, "false" device ist DOWN
	/// </summary>
	bool GetIPDeviceState(const std::string& _ethname);

	/// <summary>
	/// Gibt die aktuell vergebene IPv4-Adresse des eth-devices als _ipv4addr zurck
	/// Name des Devices in Kleinbuchstaben, z.Bsp: "eth0", "eth1.1"	
	/// </summary>
	int GetIpv4Addr(const std::string& _ethname, std::string& _ipv4addr);

	/// <summary>
	/// Gibt die aktuell vergebene Netzmaske des eth-devices als _netmask zurck
	/// Name des Devices in Kleinbuchstaben, z.Bsp: "eth0", "eth1.1"	
	/// </summary>
	int GetNetmask(const std::string& _ethname, std::string& _netmask);

	/// <summary>
	/// Gibt die aktuell vergebene Broadcast-Adresse des eth-devices als _broadcast zurck
	/// Name des Devices in Kleinbuchstaben, z.Bsp: "eth0", "eth1.1"	
	/// </summary>
	int GetBroadcast(const std::string& _ethname, std::string& _broadcast);

	/// <summary>
	/// TBD, perhaps	
	/// </summary>
	int GetIpGateway(const std::string& _ethname, std::string& _gateway);

	/// <summary>
	/// Ermittelt die IP Statistik einer Netzwerkschnittstelle
	/// </summary>
	/// <sfr>
	/// <b>enforced supported not interfered</b>
	/// </sfr>
	/// <value>
	/// Ermittelt die Anzahl der gesendeten und empfangenen Datenpakete und Bytes
	/// </value>
	/// <param name="_ethname">Name der physikalischen Netzwerkschnittstelle</param>
	/// <param name="_rxbytes">Anzahl der empfangenen Bytes</param>
	/// <param name="_txbytes">Anzahl der gesendeten Bytes</param>
	/// <param name="_rxpackets">Anzahl der empfangenen Pakete</param>
	/// <param name="_txpackets">Anzahl der gesendeten Pakete</param>
	/// <param name="_errortext">Fehlerbeschreibung im Fehlerfall</param>
	/// <returns>
	/// true fr erfolgreich, false im Fehlerfall
	/// </returns>	
	bool getIpStatistics(const std::string& _ethname, uint32_t& _rxbytes, uint32_t& _txbytes, uint32_t& _rxpackets, uint32_t& _txpackets, std::string& _errortext);
	
	/// <summary>
	/// Reset versetzt den Switch in den Default-Zustand
	/// Switch wird dann als unmanaged switch betrieben 	
	/// </summary>
	int ResetSwitch();

	/// <summary>
	/// Switch wird als managed switch konfiguriert
	/// Port1: vlan_id=1, Port2: vlan_id=2
	/// </summary>
	int ConfigureSwitch();

	/// <summary>
	/// Schaltet das GSM-Modul ein (onoff=true) bzw. ab (onoff=false)
	/// </summary>
	bool PowerGSM(bool onoff, std::string &_modell);

	/// <summary>
	/// Schaltet das WMBUS-Modul ein (onoff=true) bzw. ab (onoff=false)
	/// </summary>
	bool PowerWMBUS(bool onoff);

	/// <summary>
	/// Schaltet die +12V am Meter-Bus ein (onoff=true) bzw. ab (onoff=false)
	/// </summary>
	bool PowerMBUS12V(bool onoff);

	/// <summary>
	/// Schaltet den spezifizierten Port des Switches (1 oder 2) an (onoff=true) oder aus (onoff=false)
	/// </summary>
	bool PowerSwitchPort(unsigned int portnr, bool onoff);

	/// <summary>
	/// Alle eingehenden Datenpakete des Devices _ethname werden gedroppt (silently)
	/// equivalent zu: iptables -I INPUT 1 -i %_ethname% -j DROP
	/// </summary>	
	int IPTablesBlockInputDevice(const std::string& _ethname);
	
	/// <summary>
	/// akzeptiert Pakete fr Ziel-eth _device mit Zielport _port fr _protokoll=udp|tcp
	/// equivalent zu: iptables -I INPUT 1 -i %_device% -p %_protokoll% --dport %_port% -j ACCEPT
	/// </summary>	
	bool IPTablesOpenPort(const std::string& _device, uint16_t _port, const std::string& _protokoll, std::string& _errortext);

	/// <summary>
	/// lscht die Regel zum ffnen eines Ports aus der Kette INPUT
	/// equivalent zu: iptables -D INPUT -d %_ipv4addr% -p tcp --dport %_port% -j ACCEPT
	/// </summary>	
	bool IPTablesClosePort(const std::string& _device, uint16_t _port, const std::string& _protokoll, std::string& _errortext);
	
	/// <summary>
	/// Lscht alle Regeln der Kette INPUT
	/// equivalent zu: iptables -F INPUT
	/// </summary>	
	int IPTablesFlush();

	/// <summary>
	/// Lscht das gesamte EEPROM
	/// </summary>	
	bool EepromErase();

	/// <summary>
	/// Liest Daten aus dem EEPROM aus entsprechend den keys aus data 
	/// und liefert diese als value von data zurck
	/// 
	/// bersicht keys (first):
	/// -----------------------
	/// ver			Version, 0..999, 3-stellig
	/// hwid		Hardware-ID, 00-000..99-999, 6-stellig 
	/// swid		Software-ID, 00-000..99-999, 6-stellig
	/// hwver		Hardware-Version, 0..99999, 5-stellig
	/// osc			Oscillator, 0..2^32, 10-stellig
	/// bc			Bootcount, 0..2^32, 10-stellig
	/// mac0		MAC0, z.Bsp: 01:02:03:04:05:06, 17-stellig
	/// mac1		MAC1
	/// mac2		MAC2
	/// devid		DeviceID, z.Bsp: 0E444E540101020304, 18-stellig
	/// pdate		Production date, z.Bsp: 20121231, 8-stellig
	/// key			Schlssel, z.Bsp: 000000000000, 12-stellig
	/// oem			z.Bsp: "Dr. Neuhaus", 48-stellig
	/// wmbus		WMBUS-ID, 16-stellig
	/// mnterrinit  Mountfehler whrend Ausfhrung des InitramFS, 0..999999
	/// mnterrrun   Mountfehler zur Laufzeit, 0..999999
	/// </summary>	
	bool EepromRead(std::map<std::string, std::string>& data);
		
	/// <summary>
	/// Liest die Versionsnummer (ver) aus dem EEPROM (gepuffert)
	/// </summary>
	bool getEepromVer(std::string& _ver);

	/// <summary>
	/// Liest die Hardware-ID (hwid) aus dem EEPROM (gepuffert)
	/// </summary>
	bool getEepromHwid(std::string& _hwid);
	
	/// <summary>
	/// Liest die Software-ID (swid) aus dem EEPROM (gepuffert)
	/// </summary>
	bool getEepromSwid(std::string& _swid);

	/// <summary>
	/// Liest die Hardware-Version (hwver) aus dem EEPROM (gepuffert)
	/// </summary>
	bool getEepromHwver(std::string& _hwver);

	/// <summary>
	/// Liest die MAC-Adresse fr eth0 (mac0) aus dem EEPROM (gepuffert)
	/// </summary>
	bool getEepromMac0(std::string& _mac0);

	/// <summary>
	/// Liest die MAC-Adresse fr eth1 (mac1) aus dem EEPROM (gepuffert)
	/// </summary>
	bool getEepromMac1(std::string& _mac1);

	/// <summary>
	/// Liest die MAC-Adresse2 (mac2) aus dem EEPROM (gepuffert)
	/// </summary>
	bool getEepromMac2(std::string& _mac2);

	/// <summary>
	/// Liest die Device ID (devid) aus dem EEPROM (gepuffert)
	/// </summary>
	bool getEepromDevid(std::string& _devid);

	/// <summary>
	/// Liest das Produktionsdatum (pdate) aus dem EEPROM (gepuffert)
	/// </summary>
	bool getEepromPdate(std::string& _pdate);

	/// <summary>
	/// Liest den Eintrag des Original Equipment Manufacturer (oem) aus dem EEPROM (gepuffert)
	/// </summary>
	bool getEepromOem(std::string& _oem);

	/// <summary>
	/// Liest die WMBUS ID aus dem EEPROM (gepuffert)
	/// </summary>
	bool getEepromWmbus(std::string& _wmbus);

	/// <summary>
	/// Liest die Seriennummer aus dem EEPROM (gepuffert)
	/// </summary>
	bool getEepromSn(std::string& _sn);

	// Weitere Eintrge im EEPROM, deren Getter jedoch nicht notwendig sind:
	/*bool getEepromOsc(std::string& _osc);
	bool getEepromBc(std::string& _bc);	
	bool getEepromKey(std::string& _key);	
	bool getEepromMnterrinit(std::string& _mnterrinit);
	bool getEepromMnterrrun(std::string& _mnterrrun);
	bool getEepromRomsig(std::string& _romsig);*/
	
	/// <summary>
	/// Speichert Daten im EEPROM entsprechend den keys und values von data 	
	/// </summary>	
	bool EepromWrite(std::map<std::string, std::string>& data);

	/// <summary>
	/// Liest das Byte der Bootkonfiguration aus dem EEPROM
	/// </summary>	
	bool EepromReadBootConf(uint8_t& byte_bootregister);

	/// <summary>
	/// Schreibt das Byte der Bootkonfiguration
	/// </summary>	
	bool EepromWriteBootConf(const uint8_t byte_bootregister);

	/// <summary>
	/// Dient der Abfrage der OS-Version, Kernel-Release, installierte Updates und opkg-Pakete
	/// mode="single":
	///		Abfrage einzelner oder mehrerer Versionsnummern
	///     first="os"		-> OS-Version, z.Bsp second="DNT8209-09"
	///     first="kernel"  -> Kernel-Version und Build-Nr., z.Bsp second="Linux version 3.9.11 #16"
	///     first="%Paketname%"	-> Version eines Userland-Paketes
	///     first="smgw"		-> Version der Applikation, z.Bsp second="DNT8209-SMGW-0.1-546"
	/// mode="allpackages":
	///		Beschreibt die map mit den Paketnamen und Versionsnummern aller auf dem Target installierten Paketen
	/// mode="patches":
	///		Beschreibt die map mit den installierten Patchen
	/// </summary>	
	bool QuerySystemVersion(const std::string& _mode, std::map<std::string, std::string>& data, std::string& _error);

	/// <summary>
	/// Spezielle Updatemassnahmen jenseits des ordinren Updateverfahrens	
	/// </summary>	
	bool SystemUpdateRequest(const std::string& _pathfilename);

	/// <summary>
	/// Gibt den freien Speicher in Bytes fr das durch den Pfad und/oder Dateinamen definierte Filesystem an
	/// </summary>	
	int64_t GetFreeDiskSpace(const std::string& _pathfilename);

	/// <summary>
	/// Gibt die Gre des Filesystems in Bytes an
	/// </summary>	
	int64_t GetDiskSize(const std::string& _pathfilename);

	/// <summary>
	/// Gre des belegten Speichers eines Speicherdevices
	/// </summary>	
	int64_t GetUsedDiskSpace(const std::string& _pathfilename);

	/// <summary>
	/// Setzen des DTR-Signals, versetzt das GSM-Modul zurck in den Command-Mode und beendet pppd
	/// </summary>	
	bool SetDTR();
	
	/// <summary>
	/// Abfrage des Tamper-Kontaktes
	/// Return-Werte:
	///		0x01: Tamper OK
	///		0x02: Tamper released
	///		0x03: Tamper OK, but released in the past
	///		other: ERROR
	/// </summary>	
	int ReadTamperStatus();

	/// <summary>
	/// z.Zt. ohne Funktion
	/// </summary>	
	int TamperReboot();

	/// <summary>
	/// Gibt die PID des pppd bei aktivier Schnittstelle ppp0 zurck, im Fehlerfall: 0
	/// </summary>	
	unsigned int getPppdPid();

	/// <summary>
	/// Lst das RESET fr das WMBUS-Modul aus
	/// </summary>	
	bool ResetWMBUS();

	/// <summary>
	/// Lst das RESET fr das GSM-Modul aus
	/// </summary>	
	bool ResetGSM();

	/// <summary>
	/// Gibt den Zustand von DSR des GSM-Moduls zurck (HIGH = Command Mode, LOW = DATA-Mode)
	/// </summary>	
	bool GetDSRPegel();

	/// <summary>
	/// Gibt true zurck, sofern das angegebene Device fr die Platform existiert
	/// </summary>	
	bool EthExists(std::string &_networkDevice);
		
	/// <summary>
	/// Gibt die Spannung der Kondensatoren in mV zurck
	/// </summary>	
	unsigned int GetBattVoltage_mv();

	/// <summary>
	/// Gibt die Spannung der Kondensatoren in Prozent (100% ca. 4,22V) zurck
	/// </summary>	
	unsigned int GetBattVoltage_percent();
	
	/// <summary>
	/// Gibt den Status der Synchronisation via NTP zurck
	/// Kriterien:	- Leap indicator != 3
	///				- Ein NTP Server ist zur Zeitsynchronisation erreichbar (reach)
	///				- Der offset (Zeitdifferenz ntp<->local) liegt im limit (default = 2000 ms)
	/// </summary>		
	bool IsNTPsynchronized(uint16_t _port = 123, uint32_t _maxOffset_ms = 2000);

	/// <summary>
	/// Liest die Boot-Konfiguration aus Eeprom und CPU als Binrbyte aus
	/// </summary>	
	bool GetBootConf(uint8_t& _bootByteEeprom, uint8_t& _bootByteCPU);

	/// <summary>
	/// Setzt das Bit <bitnr> der Bootkonfiguration im EEPROM
	/// </summary>	
	bool EepromBootConfSetBit(uint8_t bitnr);

	/// <summary>
	/// Lscht das Bit <bitnr> der Bootkonfiguration im EEPROM
	/// </summary>	
	bool EepromBootConfClrBit(uint8_t bitnr);

	/// <summary>
	/// Schreibt das Byte der Bootkonfiguration in die persitenten Register der CPU
	/// </summary>	
	bool WriteBootConfToCPU(uint8_t& boot_byte);

	/// <summary>
	/// Registriert den Prozess beim imx28_power-Treiber
	/// </summary>	
	bool ActivatePowerlossDetection();

	/// <summary>
	/// Setzt die lokale Zeitzone entsprechend /usr/share/zoneinfo/, welche dann umgehend aktiv ist
	/// Bsp: hal::SetTimeZone("CET", _errortext);	equivalent: hal::SetTimeZone("Europe/Berlin", _errortext);
	/// </summary>	
	bool SetTimeZone(const std::string& _timezoneinfo, std::string& _errortext);

	/// <summary>
	/// Gibt die lokale Zeitzone entsprechend /usr/share/zoneinfo/ zurck	
	/// </summary>	
	bool GetTimeZone(std::string& _timezoneinfo, std::string& _errortext);
		
	/// <summary>
	/// Messung der Stromstrke am Meterbus
	/// </summary>
	/// <sfr>
	/// <b>enforced supported not interfered</b>
	/// </sfr>
	/// <value>
	/// Bei aktivierter 12V Ausgangsspannung des Meterbus wird 
	/// die Stromstrke bei angeschlossener Last bestimmt
	/// </value>	
	/// <returns>
	/// Rckgabe der Stromstrke in mA
	/// </returns>	
	uint32_t GetCurrent12v_mA();
	
	/// <summary>
	/// Version der SE Policy
	/// </summary>
	/// <sfr>
	/// <b>enforced supported not interfered</b>
	/// </sfr>
	/// <value>
	/// Gibt die Version fr die smgw SELINUX Policy zurck
	/// </value>
	/// <param name="_version">Versionsnummer als string</param>
	/// <param name="_errortext">Fehlerbeschreibung im Fehlerfall</param>
	/// <returns>
	/// true fr erfolgreich, false im Fehlerfall
	/// </returns>	
	//bool GetPolicyModuleVersion(std::string& _version, std::string& _errortext);

	/// <summary>
	/// Selbsttest Flashspeicher
	/// </summary>
	/// <sfr>
	/// <b>enforced supported not interfered</b>
	/// </sfr>
	/// <value>
	/// berprft den Kernellog hinsichtlich aufgetretener Fehler im UBI/UBIFS
	/// </value>	
	/// <param name="_errortext">Fehlerbeschreibung im Fehlerfall</param>
	/// <returns>
	/// true fr erfolgreicher Ausfhrung ohne gefundene Fehler, false im Fehlerfall
	/// </returns>	
	bool selfTestFlash(std::string& _errortext);

	/// <summary>
	/// Selbsttest RAM-Speicher
	/// </summary>
	/// <sfr>
	/// <b>enforced supported not interfered</b>
	/// </sfr>
	/// <value>
	/// Fhrt diverse Schreib-/Lesetests auf einen bestimmten Bereich des RAM-Speichers aus.	
	/// Folgende 7-Tests stehen zur Verfgung:
	///					- rand	(Random and Move)
	///					- sub	(Substraktion)
	///					- and	(AND-Verknpfung)
	///					- div	(Division)
	///					- mul	(Multiplikation)
	///					- or	(OR-Verknpfung)
	///					- xor	(XOR-Verknpfung)	
	/// </value>
	/// <param name="_testAlgos">Liste mit auszufhrenden Tests</param>
	/// <param name="_errortext">Fehlerbeschreibung im Fehlerfall</param>
	/// <returns>
	/// true fr erfolgreich, false im Fehlerfall
	/// </returns>	
	bool selfTestRam(const std::list<std::string>& _testAlgos, std::string& _errortext);

	/// <summary>
	/// Erstellt Liste aller laufenden Prozesse
	/// </summary>
	/// <sfr>
	/// <b>enforced supported not interfered</b>
	/// </sfr>
	/// <value>
	/// Ermittelt alle laufenden Prozesse eines Systems und gibt PID (key) und
	/// Prozessname (value) als Map zurck
	/// </value>
	/// <param name="_processList">PID und Prozessname als Map</param>
	/// <param name="_errortext">Fehlerbeschreibung im Fehlerfall</param>
	/// <returns>
	/// true fr erfolgreich, false im Fehlerfall
	/// </returns>	
	bool selfTestListRunningProcesses(std::map<std::string, std::string>& _processList, std::string& _errortext);
	
	/// <summary>
	/// Kommunikationstest mit dem securitymodule
	/// </summary>
	/// <sfr>
	/// <b>enforced supported not interfered</b>
	/// </sfr>
	/// <value>
	/// ffnet die I2c-Schnittstelle des securitymodule und fhrt eine
	/// Initialisierung durch (WakeUp + Softreset + ATR + ParameterExchange)
	/// </value>	
	/// <param name="_errortext">Fehlerbeschreibung im Fehlerfall</param>
	/// <returns>
	/// true fr erfolgreich, false im Fehlerfall
	/// </returns>	
	bool selfTestSm(std::string& _errortext);

	/// <summary>
	/// Lscht alle Prozesse eines Namens
	/// </summary>
	/// <sfr>
	/// <b>enforced supported not interfered</b>
	/// </sfr>
	/// <value>
	/// Lscht alle Prozesse eines Namens, gleiche Funktionalitt wie der Befehl "killall".
	/// Optional kann auch das Terminierungssignal explizit angegeben werden (default: SIGKILL)
	/// </value>
	/// <param name="_processName">Prozessname</param>
	/// <param name="_signal">Terminierungssignal</param>
	/// <param name="_errortext">Fehlerbeschreibung im Fehlerfall</param>
	/// <returns>
	/// true fr erfolgreich, false im Fehlerfall
	/// </returns>	
	bool KillProcessByName(const std::string& _processName, std::string& _errortext);
	bool KillProcessByName(const std::string& _processName, int32_t _signal, std::string& _errortext);
	
	// ----------------------------------------- EEPROM Definitionen --------------------------------------

	#define FCS16_LENGTH	2
	#define LENGTH_BYTE		1

	#define VERSION_ADDR	0
	#define VERSION_LENGTH	3	//	0..999

	#define HWID_ADDR		LENGTH_BYTE + VERSION_ADDR + VERSION_LENGTH + FCS16_LENGTH
	#define HWID_LENGTH		6	// 00-000..99-999

	#define SWID_ADDR	LENGTH_BYTE + HWID_ADDR + HWID_LENGTH + FCS16_LENGTH
	#define SWID_LENGTH		6	// 00-000..99-999

	#define HWVER_ADDR	LENGTH_BYTE + SWID_ADDR + SWID_LENGTH + FCS16_LENGTH
	#define HWVER_LENGTH	5	// 0..99999

	#define OSC_ADDR	LENGTH_BYTE + HWVER_ADDR + HWVER_LENGTH + FCS16_LENGTH
	#define OSC_LENGTH	10		// 0..2^32

	#define BC_ADDR	LENGTH_BYTE + OSC_ADDR + OSC_LENGTH + FCS16_LENGTH
	#define BC_LENGTH	10		// 0..2^32

	#define MAC0_ADDR	LENGTH_BYTE + BC_ADDR + BC_LENGTH + FCS16_LENGTH
	#define MAC_LENGTH	17		// typ. 01:02:03:04:05:06

	#define MAC1_ADDR	LENGTH_BYTE + MAC0_ADDR + MAC_LENGTH + FCS16_LENGTH

	#define MAC2_ADDR	LENGTH_BYTE + MAC1_ADDR + MAC_LENGTH + FCS16_LENGTH

	#define SN_ADDR		LENGTH_BYTE + MAC2_ADDR + MAC_LENGTH + FCS16_LENGTH
	#define SN_LENGTH	16		// typ. "0123456789ABCDEF"

	#define DEVID_ADDR	LENGTH_BYTE + SN_ADDR + SN_LENGTH + FCS16_LENGTH
	#define DEVID_LENGTH 18		// typ. "0E444E540101020304"

	#define PDATE_ADDR	LENGTH_BYTE + DEVID_ADDR + DEVID_LENGTH + FCS16_LENGTH
	#define PDATE_LENGTH 8		// typ. 20121231

	#define KEY_ADDR	LENGTH_BYTE + PDATE_ADDR + PDATE_LENGTH + FCS16_LENGTH
	#define KEY_LENGTH	12		// typ. "000000000000"

	#define OEM_ADDR	LENGTH_BYTE + KEY_ADDR + KEY_LENGTH + FCS16_LENGTH
	#define OEM_LENGTH	48		// typ. "Dr. Neuhaus"

	#define WMBUS_ADDR	LENGTH_BYTE + OEM_ADDR + OEM_LENGTH + FCS16_LENGTH
	#define WMBUS_LENGTH	16	// typ. "0000000000000000"

	#define MOUNTERRORINIT_ADDR LENGTH_BYTE + WMBUS_ADDR + WMBUS_LENGTH + FCS16_LENGTH 
	#define MOUNTERRORINIT_LENGTH	6		// Fehlercounter: Fehler beim Mounten einer Partition innerhalb der initramfs

	#define MOUNTERRORRUN_ADDR LENGTH_BYTE + MOUNTERRORINIT_ADDR + MOUNTERRORINIT_LENGTH + FCS16_LENGTH// Fehlercounter: Mountfehler einer Partition im laufenden Betrieb
	#define MOUNTERRORRUN_LENGTH	6		// Wert: 0..999999
	
	#define BOOTCONF_ADDR LENGTH_BYTE + MOUNTERRORRUN_ADDR + MOUNTERRORRUN_LENGTH + FCS16_LENGTH
	#define BOOTCONF_LENGTH		1	// persistierte Bootkonfiguration binr

	#define ROMSIG_ADDR BOOTCONF_ADDR + BOOTCONF_LENGTH + FCS16_LENGTH + 1 // Ein Byte Abstand zur Bootconf zwecks Erweiterung
	#define ROMSIG_LENGTH		256

	// Hashwert ganz nach hinten
	#define PWHASH_ADDR		LENGTH_BYTE + ROMSIG_LENGTH + FCS16_LENGTH
	#define PWHASH_LENGTH	32	// ?
	
	// ----------------------------------------- TAMPER Definitionen --------------------------------------

	#define UTIME	1000
	#define LOW		false
	#define HIGH	true
	#define TAMPER_CMD_READ         "00100000"		
	#define TAMPER_READ_ACK_OK		"00001111"	// Tamper OK, switch ist gedrckt und wurde noch nie zuvor released
	#define TAMPER_READ_ACK_SWITCH	"10001111"  // switch ist released -> Gehuse ist geffnet -> Alarm
	#define TAMPER_READ_ACK_EEPROM	"01001111"	// switch ist gedrckt, wurde aber in der Vergangenheit released -> stiller Alarm
	#define TAMPER_CMD_REBOOT_REQ	"00110100"	// Neustart der Tamper Firmware
	#define TAMPER_CMD_REBOOT_ACK	"00001100"	// 0xC0


}	// namespace hal


#endif	//__HAL__
