/*****************************************************************
*
* fcounter module - count frequency on GPIO 125
*
* Copyright (c) 2015 Dr. Neuhaus Telekommunikation GmbH
*
*****************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <asm/uaccess.h> 	// copy_from/to_user
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/param.h>
#include <linux/timer.h>
#include <linux/ioport.h>
#include <linux/gpio.h>
#include <linux/moduleparam.h>
#include <linux/errno.h>
#include <linux/cdev.h>
#include <linux/device.h>

// kernel module information
#define DRV_NAME "fcounter"
MODULE_LICENSE("GPL");

#define GPIO 125

// function prototypes
static void counter_exit(void);
static int32_t  counter_init(void);
static int32_t  setup_pinmux(void);
static void timer_fcounter(unsigned long arg);

// Declaration of the init and exit functions
module_init(counter_init);
module_exit(counter_exit);

// kernel timers for measurement purposes
struct timer_list timer_frequency_measurement;

static int32_t timespan = 10;	// measurement time in seconds

static uint32_t trigger = 0;
static uint32_t freqvalue_major = 0;
static uint32_t freqvalue_minor = 0;

static dev_t first;
static struct cdev fcounter_cdev;
struct class *cl = NULL;

// user space program opens this device: fopen("/dev/fcounter",.....
static int32_t counter_open(struct inode *inode, struct file *filp) 
{
	return 0;
}

static int32_t counter_release(struct inode *inode, struct file *filp) 
{
	return 0;
}

static ssize_t counter_read(struct file *filp, char *buf, size_t count, loff_t *f_pos) 
{ 
	int32_t ret = 0;
	char valuebuf[256];
	size_t ccount = count;
	size_t len = 0;

	memset(valuebuf, 0, 256);

	sprintf(valuebuf, "%d.%d", freqvalue_major, freqvalue_minor);

	len = strlen(valuebuf);

	//printk("<1>counter: strlen: %d, text: %s, major: %d, minor: %d\n", len, valuebuf, freqvalue_major, freqvalue_minor);

	if (*f_pos >= len) return 0;

	if (*f_pos + count > len) ccount = len - *f_pos;
  
	ret = copy_to_user(buf,valuebuf,ccount);
	*f_pos += ccount - ret;

	return ccount - ret;
}

struct file_operations counter_fileops = {
	read: counter_read,
	write: NULL,
	open: counter_open,
	release: counter_release
};

static void timer_fcounter(unsigned long arg)
{
	freqvalue_major = trigger / timespan;
	freqvalue_minor = trigger % timespan;
	trigger = 0;

	// restart timer
	mod_timer(&timer_frequency_measurement,jiffies+timespan*HZ);
}

// GPIO Interrupt Routine. Is called by the kernel for every edge detected on the input pin
static irqreturn_t input_irq_handler(int32_t irq, void* dev_id)
{
	trigger++;

	return IRQ_HANDLED;	// this IRQ was handled
}

static int32_t setup_pinmux(void)
{
	int32_t ret = 0;
	int32_t irqnr = gpio_to_irq(GPIO);

	if (irqnr < 0)
	{
		printk("Unable to get the irq no. from gpio device...\n");
		return -1;
	}

	//printk("IRQ GPIO %d: %d", GPIO, irqnr);
	
	ret = request_irq(irqnr, input_irq_handler, 0, "fcounter", NULL);

	//printk("rc request_irq: %d\n", ret);

	return ret;
}

static int32_t counter_init(void) 
{	
	int32_t err = 0;
	
	if (alloc_chrdev_region(&first, 0, 1, "char_dev") < 0)
	{
		return -1;
	}

	if ((cl = class_create(THIS_MODULE, "chardrv")) == NULL)
	{
		unregister_chrdev_region(first, 1);
		return -1;
	}

	if (device_create(cl, NULL, first, NULL, "fcounter") == NULL)
	{
		class_destroy(cl);
		unregister_chrdev_region(first, 1);
		return -1;
	}

	cdev_init(&fcounter_cdev, &counter_fileops);
	if (cdev_add(&fcounter_cdev, first, 1) == -1)
	{
		device_destroy(cl, first);
		class_destroy(cl);
		unregister_chrdev_region(first, 1);
		return -1;
	}	

	init_timer(&timer_frequency_measurement);					// init timer
	timer_frequency_measurement.function = timer_fcounter;		// this function is called when the timer expires
	timer_frequency_measurement.expires = jiffies + timespan*HZ;// timespan*1 second (jiffies is the current system time and HZ are the ticks per second)
	add_timer(&timer_frequency_measurement);
		
	err = setup_pinmux();
	if (err < 0) {
		printk("fcounter: failed to apply pinmux settings.\n");
		return err;
	} 
	
	printk("fcounter: driver loaded sucessfully\n");	

	return 0;
}

static void counter_exit(void) 
{
	// remove the timers
	del_timer(&timer_frequency_measurement);
	
	free_irq(GPIO, NULL);
	
	device_destroy(cl, first);
	class_destroy(cl);
	unregister_chrdev_region(first, 1);
	
	//printk("<1> counter_driver: driver unloaded\n");	
}