#include <linux/ctype.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <asm/io.h>

#include <linux/version.h>
#if LINUX_VERSION_CODE <= KERNEL_VERSION(3,9,11)
#include <mach/mx28.h>
#else
#include "mx28.h"
#endif

MODULE_AUTHOR("B.Bauch");
MODULE_DESCRIPTION("dump iMX28 Registers");
MODULE_LICENSE("GPL");

static char reg_name[10] = "";
module_param_string(reg, reg_name, sizeof(reg_name), S_IRUGO);
MODULE_PARM_DESC(reg, "iMX28 register name: PINCTRL, DCP, ...");

static char *strtoupper(char *text)
{
	size_t i, len = strlen(text);
	for (i=0; i < len; i++)
		text[i] = toupper(text[i]);
	return text;
}

static int printMemoryMap(const char *text, const unsigned int base, const unsigned int size, const unsigned int regsize, const unsigned int offset)
{
	unsigned int pos = 0;
	unsigned int count = 0;
	unsigned int len = size * offset;

	if (offset < regsize)
		return -EFAULT;

	printk("---------------------------------------------------------------------------------------\n");
	printk(text);
	printk("\n");
	printk("---------------------------------------------------------------------------------------\n");

	while (pos < len)
	{
		void *temp_reg = ioremap(base + pos, regsize);
		if (!temp_reg)
		{
			printk("\nioremap failed at: 0x%X\n", base + pos);
			return -EFAULT;
		}
		else
		{
			printk("0x%X:0x%.8X ", base + pos, __raw_readl(temp_reg));
			iounmap(temp_reg);
		}
		pos += offset;
		if (++count == 4)
		{
			count = 0;
			printk("\n");
		}
	}
	if (count)
		printk("\n");

	return 0;
}

static int __init mod_init(void)
{
	printk("-------------------------------------------------------------------- iMX28 memory map -\n");

	if (0 == strcmp(strtoupper(reg_name), "ICOLL"))
	{
		printMemoryMap("ICOLL", MX28_ICOLL_BASE_ADDR, 287, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "HSADC"))
	{
		printMemoryMap("HSADC", MX28_HSADC_BASE_ADDR, 12, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "PERFMON"))
	{
		printMemoryMap("PERFMON", MX28_PERFMON_BASE_ADDR, 12, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "BCH"))
	{
		printMemoryMap("BCH", MX28_BCH_BASE_ADDR, 23, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "GPMI"))
	{
		printMemoryMap("GPMI", MX28_GPMI_BASE_ADDR, 14, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "SSP"))
	{
		printMemoryMap("SSP0", MX28_SSP0_BASE_ADDR, 20, 4, 16);
		printMemoryMap("SSP1", MX28_SSP1_BASE_ADDR, 20, 4, 16);
		printMemoryMap("SSP2", MX28_SSP2_BASE_ADDR, 20, 4, 16);
		printMemoryMap("SSP3", MX28_SSP3_BASE_ADDR, 20, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "PINCTRL"))
	{
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR, 1, 4, 16);
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR + 0x0100, 14, 4, 16);
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR + 0x0300, 20, 4, 16);
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR + 0x0600, 7, 4, 16);
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR + 0x0700, 5, 4, 16);
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR + 0x0900, 5, 4, 16);
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR + 0x0B00, 5, 4, 16);
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR + 0x1000, 5, 4, 16);
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR + 0x1100, 5, 4, 16);
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR + 0x1200, 5, 4, 16);
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR + 0x1300, 5, 4, 16);
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR + 0x1400, 5, 4, 16);
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR + 0x1A40, 1, 4, 16);
		printMemoryMap("PINCTRL", MX28_PINCTRL_BASE_ADDR + 0x1B80, 1, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "DIGCTL"))
	{
		printMemoryMap("DIGCTL", MX28_DIGCTL_BASE_ADDR, 96, 4, 16); //include some undocumented registers
	}
	else if (0 == strcmp(strtoupper(reg_name), "DCP"))
	{
		printMemoryMap("DCP", MX28_DCP_BASE_ADDR, 68, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "PXP"))
	{
		printMemoryMap("PXP", MX28_PXP_BASE_ADDR, 64, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "OCOTP"))
	{
		printMemoryMap("OCOTP", MX28_OCOTP_BASE_ADDR, 43, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "LCDIF"))
	{
		printMemoryMap("LCDIF", MX28_LCDIF_BASE_ADDR, 32, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "CLKCTRL"))
	{
		printMemoryMap("CLKCTRL", MX28_CLKCTRL_BASE_ADDR, 33, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "SAIF"))
	{
		printMemoryMap("SAIF0", MX28_SAIF0_BASE_ADDR, 4, 4, 16);
		printMemoryMap("SAIF1", MX28_SAIF1_BASE_ADDR, 4, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "POWER"))
	{
		printMemoryMap("POWER", MX28_POWER_BASE_ADDR, 24, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "LRADC"))
	{
		printMemoryMap("LRADC", MX28_LRADC_BASE_ADDR, 24, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "SPDIF"))
	{
		printMemoryMap("SPDIF", MX28_SPDIF_BASE_ADDR, 7, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "RTC"))
	{
		printMemoryMap("RTC", MX28_RTC_BASE_ADDR, 14, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "I2C"))
	{
		printMemoryMap("I2C0", MX28_I2C0_BASE_ADDR, 14, 4, 16);
		printMemoryMap("I2C1", MX28_I2C1_BASE_ADDR, 14, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "PWM"))
	{
		printMemoryMap("PWM", MX28_PWM_BASE_ADDR, 18, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "TIMROT"))
	{
		printMemoryMap("TIMROT", MX28_TIMROT_BASE_ADDR, 19, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "AUART"))
	{
		printMemoryMap("AUART0", MX28_AUART0_BASE_ADDR, 11, 4, 16);
		printMemoryMap("AUART1", MX28_AUART1_BASE_ADDR, 11, 4, 16);
		printMemoryMap("AUART2", MX28_AUART2_BASE_ADDR, 11, 4, 16);
		printMemoryMap("AUART3", MX28_AUART3_BASE_ADDR, 11, 4, 16);
		printMemoryMap("AUART4", MX28_AUART4_BASE_ADDR, 11, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "DUART"))
	{
		printMemoryMap("DUART", MX28_DUART_BASE_ADDR, 19, 4, 4);			//include some undocumented registers
	}
	else if (0 == strcmp(strtoupper(reg_name), "ENET"))
	{
		printMemoryMap("ENET_MAC0", MX28_ENET_MAC0_BASE_ADDR, 105, 4, 16);
		printMemoryMap("ENET_MAC1", MX28_ENET_MAC1_BASE_ADDR + 0x4000, 105, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "USBPHY"))
	{
		printMemoryMap("USBPHY0", MX28_USBPHY0_BASE_ADDR, 10, 4, 16);
		printMemoryMap("USBPHY1", MX28_USBPHY1_BASE_ADDR, 10, 4, 16);
	}
	else if (0 == strcmp(strtoupper(reg_name), "USBCTRL"))
	{
		printMemoryMap("USBCTRL0", MX28_USBCTRL0_BASE_ADDR, 120, 4, 4);
		printMemoryMap("USBCTRL1", MX28_USBCTRL1_BASE_ADDR, 120, 4, 4);
	}
	else
	{
		printk("usage: insmod imx28_dump reg=name\n");
		printk("with name of: \n");
		printk("  ICOLL, HSADC, PERFMON, BCH, GPMI, SSP, PINCTRL, DIGCTL, DCP, \n");
		printk("  PXP, OCOTP, LCDIF, CLKCTRL, SAIF, POWER, LRADC, SPDIF, RTC, \n");
		printk("  I2C, PWM, TIMROT, AUART, DUART, ENET, USBPHY, USBCTRL\n");
	}

	printk("--------------------------------------------------------------------------------- bye -\n");

	return -EPERM;
}

static void __exit mod_exit(void)
{
	//nothing to do
}

module_init(mod_init);
module_exit(mod_exit);
