// ************************************************************

#include <linux/delay.h>
#include <linux/err.h>
#include <linux/mutex.h>
#include <asm/processor.h>	/* for cpu_relax() */
#include <asm/errno.h>

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/platform_device.h>
#include <linux/device.h>
#include <linux/cdev.h>
#include <linux/string.h>
#include <linux/poll.h>
#include <linux/sched.h>

#include <mach/mx28.h>

// ************************************************************

#define DRIVER_NAME		"imx28_regs"
#define DRIVER_VERSION	"0.0.1"

static dev_t first;
static struct cdev userio_cdev;
static struct class *userio_class;

// ************************************************************

static unsigned int pwm3_clock = 0;
#define PWM_CHANNEL 3
#include <linux/pwm.h>
static struct pwm_device *pwm = NULL;

enum
{
	PWM3_DISABLED,
	PWM3_32KHZ,
	PWM3_24MHZ,
};

#ifndef CALCULATE_PWM_PERIOD
// idea stolen from linux-3.9.11/drivers/pwm/pwm-mxs.c
#define PWM_PERIOD0             0x20
#define BM_PWM_PERIODn_MATT_SEL 0x01000000
#define BM_PWM_PERIODn_MATT     0x00800000
struct mxs_pwm_chip {
	struct pwm_chip chip;
	struct device *dev;
	struct clk *clk;
	void __iomem *base;
};
#define to_mxs_pwm_chip(_chip) container_of(_chip, struct mxs_pwm_chip, chip)
#endif

static void setPwm3clock(unsigned int clock)
{
	//set clock on PWM3 using pwm kernel device
	switch (clock)
	{
		default:
		case PWM3_DISABLED:
			if (pwm)
			{
#ifdef CALCULATE_PWM_PERIOD
				int rc = pwm_config(pwm, 0, 1000000);
#else
				struct pwm_chip	*chip = pwm->chip;
				struct mxs_pwm_chip *mxs = to_mxs_pwm_chip(chip);
				writel(0, mxs->base + PWM_PERIOD0 + PWM_CHANNEL * 0x20);
#endif
				pwm_disable(pwm);
				pwm_free(pwm);
				pwm = NULL;
			}
			printk(KERN_INFO "%s: disabled\n", __func__);
			break;
		case PWM3_32KHZ:
		case PWM3_24MHZ:
			if (pwm == NULL)
			{
				pwm = pwm_request(PWM_CHANNEL, "pwm");
				if (IS_ERR(pwm))
					pwm = NULL;
			}
			if (pwm)
			{
#ifdef CALCULATE_PWM_PERIOD
				int rc, period = (clock == PWM3_24MHZ) ? (uint32_t)(1000.0/24.0 + 0.5) : (uint32_t)(1000000.0/32.0 + 0.5);
				pwm_enable(pwm);
				rc = pwm_config(pwm, period >> 1, period);
				printk(KERN_INFO "%s: enable %s pwm clock with period=%dns, rc=%d\n", __func__, (clock == PWM3_24MHZ) ? "24MHz":"32KHz", period, rc);
#else
				struct pwm_chip	*chip = pwm->chip;
				struct mxs_pwm_chip *mxs = to_mxs_pwm_chip(chip);
				// make sure to turn on the clock before writing the register
				pwm_enable(pwm);
				if (clock == PWM3_24MHZ)
					writel(BM_PWM_PERIODn_MATT | BM_PWM_PERIODn_MATT_SEL, mxs->base + PWM_PERIOD0 + PWM_CHANNEL * 0x20);
				else
					writel(BM_PWM_PERIODn_MATT, mxs->base + PWM_PERIOD0 + PWM_CHANNEL * 0x20);
				printk(KERN_INFO "%s: enable %s pwm clock\n", __func__, (clock == PWM3_24MHZ) ? "24MHz":"32KHz");
#endif
			}
			else
				printk(KERN_INFO "%s: failed\n", __func__);
			break;
	}
}

// ************************************************************

#define RTC_BASE				MX28_IO_ADDRESS(MX28_RTC_BASE_ADDR)
#define RTC_PERSISTENT1			(RTC_BASE + 0x070)
#define BF_ROM_SECONDARY_BOOT	0x00000002
#define RTC_PERSISTENT2			(RTC_BASE + 0x080)
#define BM_USER_BIT_0			0x00000001
#define BM_USER_BIT_1			0x00000002
#define BM_USER_BIT_2			0x00000004
#define BM_USER_BIT_3			0x00000008
#define BM_USER_BIT_4			0x00000010
#define BM_USER_BIT_5			0x00000020
#define BM_USER_BIT_6			0x00000040
#define BM_USER_BIT_7			0x00000080

static void set_ROM_SECONDARY_BOOT(int flag)
{
	uint32_t val = readl_relaxed(RTC_PERSISTENT1);
	if (flag)
		val |= BF_ROM_SECONDARY_BOOT;
	else
		val &= ~BF_ROM_SECONDARY_BOOT;
	writel_relaxed(val, RTC_PERSISTENT1);
}

static int get_ROM_SECONDARY_BOOT(void)
{
	return (((readl_relaxed(RTC_PERSISTENT1) & BF_ROM_SECONDARY_BOOT) == BF_ROM_SECONDARY_BOOT) ? 1:0);
}

static void set_USER_BIT(uint32_t mask, int flag)
{
	uint32_t val = readl_relaxed(RTC_PERSISTENT2);
	if (flag)
		val |= mask;
	else
		val &= ~mask;
	writel_relaxed(val, RTC_PERSISTENT2);
}

static int get_USER_BIT(uint32_t mask)
{
	return (((readl_relaxed(RTC_PERSISTENT2) & mask) == mask) ? 1:0);
}

// ************************************************************

#define POWER_BASE						MX28_IO_ADDRESS(MX28_POWER_BASE_ADDR)
#define POWER_BATTMONITOR				(POWER_BASE + 0x0e0)
#define BP_POWER_BATTMONITOR_BATT_VAL	16
#define BM_POWER_BATTMONITOR_BATT_VAL	0x03FF0000
#define BATT_VOLTAGE_8_MV				8

static uint16_t get_BATT_VAL(void)
{
	// Get the raw result of battery measurement
	uint32_t u16BattVolt = (readl(POWER_BATTMONITOR) & BM_POWER_BATTMONITOR_BATT_VAL) >> BP_POWER_BATTMONITOR_BATT_VAL;
	// Adjust for 8-mV LSB resolution and return
	return u16BattVolt * BATT_VOLTAGE_8_MV;
}

// ************************************************************

#define OCOTP_WORD_OFFSET		0x20
#define OCOTP_WORD_COUNT		0x20

#if 0

/*
 * intermodul communication
 *
 * mxs_get_ocotp and ocotp_words should be exported inside linux-3.9.11/arch/arm/mach-mxs/ocotp.c
 *
 * EXPORT_SYMBOL( ocotp_words );
 * EXPORT_SYMBOL_GPL( mxs_get_ocotp );
 *
 * But it does not work...
 */
#include <mach/common.h>

#else
/*
 * linux-3.9.11/arch/arm/mach-mxs/ocotp.c
 */
#define BM_OCOTP_CTRL_BUSY		(1 << 8)
#define BM_OCOTP_CTRL_ERROR		(1 << 9)
#define BM_OCOTP_CTRL_RD_BANK_OPEN	(1 << 12)

static DEFINE_MUTEX(ocotp_mutex);
static u32 ocotp_words[OCOTP_WORD_COUNT];

const u32 *mxs_get_ocotp(void)
{
	void __iomem *ocotp_base = MXS_IO_ADDRESS(MXS_OCOTP_BASE_ADDR);
	int timeout = 0x400;
	size_t i;
	static int once = 0;

	if (once)
		return ocotp_words;

	mutex_lock(&ocotp_mutex);

	/*
	 * clk_enable(hbus_clk) for ocotp can be skipped
	 * as it must be on when system is running.
	 */

	/* try to clear ERROR bit */
	__mxs_clrl(BM_OCOTP_CTRL_ERROR, ocotp_base);

	/* check both BUSY and ERROR cleared */
	while ((__raw_readl(ocotp_base) &
		(BM_OCOTP_CTRL_BUSY | BM_OCOTP_CTRL_ERROR)) && --timeout)
		cpu_relax();

	if (unlikely(!timeout))
		goto error_unlock;

	/* open OCOTP banks for read */
	__mxs_setl(BM_OCOTP_CTRL_RD_BANK_OPEN, ocotp_base);

	/* approximately wait 32 hclk cycles */
	udelay(1);

	/* poll BUSY bit becoming cleared */
	timeout = 0x400;
	while ((__raw_readl(ocotp_base) & BM_OCOTP_CTRL_BUSY) && --timeout)
		cpu_relax();

	if (unlikely(!timeout))
		goto error_unlock;

	for (i = 0; i < OCOTP_WORD_COUNT; i++)
		ocotp_words[i] = __raw_readl(ocotp_base + OCOTP_WORD_OFFSET +
						i * 0x10);

	/* close banks for power saving */
	__mxs_clrl(BM_OCOTP_CTRL_RD_BANK_OPEN, ocotp_base);

	once = 1;

	mutex_unlock(&ocotp_mutex);

	return ocotp_words;

error_unlock:
	mutex_unlock(&ocotp_mutex);
	pr_err("%s: timeout in reading OCOTP\n", __func__);
	return NULL;
}
#endif

static int get_ocotp(void)
{
	int i;
	const uint32_t *ocotp = mxs_get_ocotp();
	for (i=0; i<OCOTP_WORD_COUNT; i++)
		printk(KERN_INFO "bank %d-%d: %08x\n", (i >> 3), (i & 0x07), ocotp[i]);
	return 0;
}

// show the driver version in "/sys/devices/platform/imx28_regs/driver/version"
static ssize_t driver_version(struct device_driver *driver, char *buffer)
{
	strcpy(buffer, DRIVER_VERSION);
	strcat(buffer, " (");
	strcat(buffer, __DATE__);
	strcat(buffer, " - ");
	strcat(buffer, __TIME__);
	strcat(buffer, ")");
	return strlen(buffer) + 1;
}
static DRIVER_ATTR(version, S_IRUGO, driver_version, NULL);

// read/write status of ROM_SECONDARY_BOOT in "/sys/devices/platform/imx28_regs/rom_secondary_boot"
static ssize_t show_rom_secondary_boot(struct device *dev, struct device_attribute *attr, char *buffer)
{
	sprintf(buffer, "%d", (get_ROM_SECONDARY_BOOT()) ? 1 : 0);
	return (strlen(buffer) + 1);
}
static ssize_t store_rom_secondary_boot(struct device *dev, struct device_attribute *attr, const char *buffer, size_t count)
{
	set_ROM_SECONDARY_BOOT((simple_strtoul(buffer, NULL, 0)) ? 1:0);
	return count;
}
static DEVICE_ATTR(rom_secondary_boot, S_IRUGO | S_IWUGO, show_rom_secondary_boot, store_rom_secondary_boot);

// read/write status in "/sys/devices/platform/imx28_regs/user_bit_0"
static ssize_t show_user_0(struct device *dev, struct device_attribute *attr, char *buffer)
{
	sprintf(buffer, "%d", (get_USER_BIT(BM_USER_BIT_0)) ? 1 : 0);
	return (strlen(buffer) + 1);
}
static ssize_t store_user_0(struct device *dev, struct device_attribute *attr, const char *buffer, size_t count)
{
	set_USER_BIT(BM_USER_BIT_0, (simple_strtoul(buffer, NULL, 0)) ? 1:0);
	return count;
}
static DEVICE_ATTR(user_bit_0, S_IRUGO | S_IWUGO, show_user_0, store_user_0);

// read/write status in "/sys/devices/platform/imx28_regs/user_bit_1"
static ssize_t show_user_1(struct device *dev, struct device_attribute *attr, char *buffer)
{
	sprintf(buffer, "%d", (get_USER_BIT(BM_USER_BIT_1)) ? 1 : 0);
	return (strlen(buffer) + 1);
}
static ssize_t store_user_1(struct device *dev, struct device_attribute *attr, const char *buffer, size_t count)
{
	set_USER_BIT(BM_USER_BIT_1, (simple_strtoul(buffer, NULL, 0)) ? 1:0);
	return count;
}
static DEVICE_ATTR(user_bit_1, S_IRUGO | S_IWUGO, show_user_1, store_user_1);

// read/write status in "/sys/devices/platform/imx28_regs/user_bit_2"
static ssize_t show_user_2(struct device *dev, struct device_attribute *attr, char *buffer)
{
	sprintf(buffer, "%d", (get_USER_BIT(BM_USER_BIT_2)) ? 1 : 0);
	return (strlen(buffer) + 1);
}
static ssize_t store_user_2(struct device *dev, struct device_attribute *attr, const char *buffer, size_t count)
{
	set_USER_BIT(BM_USER_BIT_2, (simple_strtoul(buffer, NULL, 0)) ? 1:0);
	return count;
}
static DEVICE_ATTR(user_bit_2, S_IRUGO | S_IWUGO, show_user_2, store_user_2);

// read/write status in "/sys/devices/platform/imx28_regs/user_bit_3"
static ssize_t show_user_3(struct device *dev, struct device_attribute *attr, char *buffer)
{
	sprintf(buffer, "%d", (get_USER_BIT(BM_USER_BIT_3)) ? 1 : 0);
	return (strlen(buffer) + 1);
}
static ssize_t store_user_3(struct device *dev, struct device_attribute *attr, const char *buffer, size_t count)
{
	set_USER_BIT(BM_USER_BIT_3, (simple_strtoul(buffer, NULL, 0)) ? 1:0);
	return count;
}
static DEVICE_ATTR(user_bit_3, S_IRUGO | S_IWUGO, show_user_3, store_user_3);

// read/write status in "/sys/devices/platform/imx28_regs/user_bit_4"
static ssize_t show_user_4(struct device *dev, struct device_attribute *attr, char *buffer)
{
	sprintf(buffer, "%d", (get_USER_BIT(BM_USER_BIT_4)) ? 1 : 0);
	return (strlen(buffer) + 1);
}
static ssize_t store_user_4(struct device *dev, struct device_attribute *attr, const char *buffer, size_t count)
{
	set_USER_BIT(BM_USER_BIT_4, (simple_strtoul(buffer, NULL, 0)) ? 1:0);
	return count;
}
static DEVICE_ATTR(user_bit_4, S_IRUGO | S_IWUGO, show_user_4, store_user_4);

// read/write status in "/sys/devices/platform/imx28_regs/user_bit_5"
static ssize_t show_user_5(struct device *dev, struct device_attribute *attr, char *buffer)
{
	sprintf(buffer, "%d", (get_USER_BIT(BM_USER_BIT_5)) ? 1 : 0);
	return (strlen(buffer) + 1);
}
static ssize_t store_user_5(struct device *dev, struct device_attribute *attr, const char *buffer, size_t count)
{
	set_USER_BIT(BM_USER_BIT_5, (simple_strtoul(buffer, NULL, 0)) ? 1:0);
	return count;
}
static DEVICE_ATTR(user_bit_5, S_IRUGO | S_IWUGO, show_user_5, store_user_5);

// read/write status in "/sys/devices/platform/imx28_regs/user_bit_6"
static ssize_t show_user_6(struct device *dev, struct device_attribute *attr, char *buffer)
{
	sprintf(buffer, "%d", (get_USER_BIT(BM_USER_BIT_6)) ? 1 : 0);
	return (strlen(buffer) + 1);
}
static ssize_t store_user_6(struct device *dev, struct device_attribute *attr, const char *buffer, size_t count)
{
	set_USER_BIT(BM_USER_BIT_6, (simple_strtoul(buffer, NULL, 0)) ? 1:0);
	return count;
}
static DEVICE_ATTR(user_bit_6, S_IRUGO | S_IWUGO, show_user_6, store_user_6);

// read/write status in "/sys/devices/platform/imx28_regs/user_bit_7"
static ssize_t show_user_7(struct device *dev, struct device_attribute *attr, char *buffer)
{
	sprintf(buffer, "%d", (get_USER_BIT(BM_USER_BIT_7)) ? 1 : 0);
	return (strlen(buffer) + 1);
}
static ssize_t store_user_7(struct device *dev, struct device_attribute *attr, const char *buffer, size_t count)
{
	set_USER_BIT(BM_USER_BIT_7, (simple_strtoul(buffer, NULL, 0)) ? 1:0);
	return count;
}
static DEVICE_ATTR(user_bit_7, S_IRUGO | S_IWUGO, show_user_7, store_user_7);

// read/write status in "/sys/devices/platform/imx28_regs/pwm3"
static ssize_t show_pwm3(struct device *dev, struct device_attribute *attr, char *buffer)
{
	sprintf(buffer, "%d", pwm3_clock);
	return (strlen(buffer) + 1);
}
static ssize_t store_pwm3(struct device *dev, struct device_attribute *attr, const char *buffer, size_t count)
{
	pwm3_clock = simple_strtoul(buffer, NULL, 0);
	setPwm3clock(pwm3_clock);
	return count;
}
static DEVICE_ATTR(pwm3, S_IRUGO | S_IWUGO, show_pwm3, store_pwm3);

// read status in "/sys/devices/platform/imx28_regs/battery"
static ssize_t show_battery(struct device *dev, struct device_attribute *attr, char *buffer)
{
	sprintf(buffer, "%d", get_BATT_VAL());
	return (strlen(buffer) + 1);
}
static DEVICE_ATTR(battery, S_IRUGO, show_battery, NULL);

// read status in "/sys/devices/platform/imx28_regs/ocotp"
static ssize_t show_ocotp(struct device *dev, struct device_attribute *attr, char *buffer)
{
	sprintf(buffer, "%d", get_ocotp());
	return (strlen(buffer) + 1);
}
static DEVICE_ATTR(ocotp, S_IRUGO, show_ocotp, NULL);

static int file_open(struct inode *inode, struct file *filp)
{
	printk(KERN_DEBUG "%s: process %i (%s) ref %ld\n", __func__, current->pid, current->comm, module_refcount (THIS_MODULE));
	return 0;
}

static int file_release (struct inode *inode, struct file *filp)
{
	printk(KERN_DEBUG "%s: process %i (%s) ref %ld\n", __func__, current->pid, current->comm, module_refcount (THIS_MODULE));
	return 0;
}

static const struct file_operations fops =
{
	.owner = THIS_MODULE,
	.open = file_open,
	.release = file_release,
};

static void mod_release(struct device *dev)
{
	printk(KERN_INFO "%s: release character device %s\n", __func__, DRIVER_NAME);
	setPwm3clock(PWM3_DISABLED);
}

struct platform_device pdev = {
	.name = DRIVER_NAME,
	.id = 0,
	.dev = {
		.release = mod_release,
	}
};

static struct device_driver userio_drv = {
	.name = DRIVER_NAME,
	.bus = &platform_bus_type,
};

static int __init initialize_cdev(void)
{
	if (alloc_chrdev_region(&first, 0, 1, DRIVER_NAME) < 0)
	{
		printk(KERN_ALERT "%s: alloc_chrdev_region() failed\n", __func__);
		return -EIO;
	}

	cdev_init(&userio_cdev, &fops);
	userio_cdev.owner = THIS_MODULE;

	if (cdev_add(&userio_cdev, first, 1) < 0)
	{
		printk(KERN_ALERT "%s: cdev_add() failed\n", __func__);
		unregister_chrdev_region(first, 1);
		return -EIO;
	}
	return 0;
}

static int __init initialize_class(void)
{
	userio_class = class_create(THIS_MODULE, DRIVER_NAME);
	if (IS_ERR(userio_class))
	{
		printk(KERN_ALERT "%s: class_create(..., %s) failed\n", __func__, DRIVER_NAME);
		return -EIO;
	}
	if (!device_create(userio_class, NULL, first, NULL, "userio"))
	{
		printk(KERN_ALERT "%s: device_create(..., %s) failed\n", __func__, DRIVER_NAME);
		class_destroy(userio_class);
		return -EIO;
	}
	return 0;
}

static int __init mod_init(void)
{
	if (initialize_cdev() < 0)
		goto fail_1;

	if (initialize_class() < 0)
		goto fail_2;

	//register the driver
	if (0 == driver_register(&userio_drv))
	{
		//register the device
		if (0 == platform_device_register(&pdev))
		{
			//now tie them together
			pdev.dev.driver = &userio_drv;
			//links the driver to the device
			if (0 == device_bind_driver(&pdev.dev))
			{
				//create attribute file "version"
				if (0 == driver_create_file(&userio_drv, &driver_attr_version))
				{
					if (0 == device_create_file(&pdev.dev, &dev_attr_rom_secondary_boot))
					{
						if (0 == device_create_file(&pdev.dev, &dev_attr_user_bit_0))
						{
							if (0 == device_create_file(&pdev.dev, &dev_attr_user_bit_1))
							{
								if (0 == device_create_file(&pdev.dev, &dev_attr_user_bit_2))
								{
									if (0 == device_create_file(&pdev.dev, &dev_attr_user_bit_3))
									{
										if (0 == device_create_file(&pdev.dev, &dev_attr_user_bit_4))
										{
											if (0 == device_create_file(&pdev.dev, &dev_attr_user_bit_5))
											{
												if (0 == device_create_file(&pdev.dev, &dev_attr_user_bit_6))
												{
													if (0 == device_create_file(&pdev.dev, &dev_attr_user_bit_7))
													{
														if (0 == device_create_file(&pdev.dev, &dev_attr_pwm3))
														{
															if (0 == device_create_file(&pdev.dev, &dev_attr_battery))
															{
																if (0 == device_create_file(&pdev.dev, &dev_attr_ocotp))
																{
																	printk(KERN_INFO "%s: registering character device %s (major=%d, minor=%d)\n", __func__, DRIVER_NAME, MAJOR (first), MINOR (first));
																	return 0;

																	device_remove_file( &pdev.dev, &dev_attr_ocotp );
																}
																else
																	printk(KERN_ERR "%s: can't create file ocotp\n", __func__);

																device_remove_file( &pdev.dev, &dev_attr_battery );
															}
															else
																printk(KERN_ERR "%s: can't create file battery\n", __func__);

															device_remove_file( &pdev.dev, &dev_attr_pwm3 );
														}
														else
															printk(KERN_ERR "%s: can't create file pwm3\n", __func__);

														device_remove_file( &pdev.dev, &dev_attr_user_bit_7 );
													}
													else
														printk(KERN_ERR "%s: can't create file user_bit_7\n", __func__);

													device_remove_file( &pdev.dev, &dev_attr_user_bit_6 );
												}
												else
													printk(KERN_ERR "%s: can't create file user_bit_6\n", __func__);

												device_remove_file( &pdev.dev, &dev_attr_user_bit_5 );
											}
											else
												printk(KERN_ERR "%s: can't create file user_bit_5\n", __func__);

											device_remove_file( &pdev.dev, &dev_attr_user_bit_4 );
										}
										else
											printk(KERN_ERR "%s: can't create file user_bit_4\n", __func__);

										device_remove_file( &pdev.dev, &dev_attr_user_bit_3 );
									}
									else
										printk(KERN_ERR "%s: can't create file user_bit_3\n", __func__);

									device_remove_file( &pdev.dev, &dev_attr_user_bit_2 );
								}
								else
									printk(KERN_ERR "%s: can't create file user_bit_2\n", __func__);

								device_remove_file( &pdev.dev, &dev_attr_user_bit_1 );
							}
							else
								printk(KERN_ERR "%s: can't create file user_bit_1\n", __func__);

							device_remove_file( &pdev.dev, &dev_attr_user_bit_0 );
						}
						else
							printk(KERN_ERR "%s: can't create file user_bit_1\n", __func__);

						device_remove_file( &pdev.dev, &dev_attr_rom_secondary_boot );
					}
					else
						printk(KERN_ERR "%s: can't create file rom_secondary_boot\n", __func__);

					driver_remove_file( &userio_drv, &driver_attr_version );
				}
				else
					printk(KERN_ERR "%s: can't create file version\n", __func__);

				device_release_driver( &pdev.dev );
			}
			else
				printk(KERN_ERR "%s: can't link driver to the device\n", __func__);

			platform_device_unregister( &pdev );
		}
		else
			printk(KERN_ERR "%s: can't register the device\n", __func__);

		driver_unregister(&userio_drv);
	}
	else
		printk(KERN_ERR "%s: can't register the driver\n", __func__);

	device_destroy(userio_class, first);
	class_destroy(userio_class);

fail_2:
	cdev_del(&userio_cdev);
	unregister_chrdev_region(first, 1);

fail_1:
	return -1;
}

static void __exit mod_exit(void)
{
	device_remove_file( &pdev.dev, &dev_attr_ocotp );
	device_remove_file( &pdev.dev, &dev_attr_battery );
	device_remove_file( &pdev.dev, &dev_attr_pwm3 );
	device_remove_file( &pdev.dev, &dev_attr_user_bit_7 );
	device_remove_file( &pdev.dev, &dev_attr_user_bit_6 );
	device_remove_file( &pdev.dev, &dev_attr_user_bit_5 );
	device_remove_file( &pdev.dev, &dev_attr_user_bit_4 );
	device_remove_file( &pdev.dev, &dev_attr_user_bit_3 );
	device_remove_file( &pdev.dev, &dev_attr_user_bit_2 );
	device_remove_file( &pdev.dev, &dev_attr_user_bit_1 );
	device_remove_file( &pdev.dev, &dev_attr_user_bit_0 );
	device_remove_file( &pdev.dev, &dev_attr_rom_secondary_boot );
	driver_remove_file( &userio_drv, &driver_attr_version );

	device_release_driver( &pdev.dev );
	platform_device_unregister( &pdev );
	driver_unregister(&userio_drv);

	device_destroy(userio_class, first);
	class_destroy(userio_class);

	cdev_del(&userio_cdev);
	unregister_chrdev_region(first, 1);
}

module_init( mod_init );
module_exit( mod_exit );

MODULE_AUTHOR("B.Bauch");
MODULE_DESCRIPTION("userio driver for some imx28 registers");
MODULE_LICENSE("GPL");
MODULE_VERSION("0.1");
