/*****************************************************************
 *
 * EEPROM - Utilities
 *
 * Copyright (c) 2011 Birger Bauch, Dr. Neuhaus Telekommunikation GmbH
 *
 *****************************************************************/

#define DEVICE_I2C_EEPROM	"/sys/bus/i2c/devices/0-0050/eeprom"
#define SIZEOF_I2C_EEPROM	512

#define FCS16_LENGTH	2

#define VERSION_ADDR	0
#define VERSION_LENGTH	1	//uint8_t

#define OSC_ADDR		VERSION_ADDR + VERSION_LENGTH + FCS16_LENGTH
#define OSC_LENGTH		4	//uint32_t

#define BC_ADDR			OSC_ADDR + OSC_LENGTH + FCS16_LENGTH
#define BC_LENGTH		4	//uint32_t

#define HWID_ADDR		BC_ADDR + BC_LENGTH + FCS16_LENGTH
#define HWID_LENGTH		6

#define SWID_ADDR		HWID_ADDR + HWID_LENGTH + FCS16_LENGTH
#define SWID_LENGTH		6

#define SN_ADDR			SWID_ADDR + SWID_LENGTH + FCS16_LENGTH
#define SN_LENGTH		16

#define PDATE_ADDR		SN_ADDR + SN_LENGTH + FCS16_LENGTH
#define PDATE_LENGTH	4+2+2

#define HWVER_ADDR		PDATE_ADDR + PDATE_LENGTH + FCS16_LENGTH
#define HWVER_LENGTH	5

#define MAC0_ADDR		HWVER_ADDR + HWVER_LENGTH + FCS16_LENGTH
#define MAC_LENGTH		17
#define MAC1_ADDR		MAC0_ADDR + MAC_LENGTH + FCS16_LENGTH

#define KEY_ADDR		MAC1_ADDR + MAC_LENGTH + FCS16_LENGTH
#define KEY_LENGTH		12

#define FLAG_ADDR		KEY_ADDR + KEY_LENGTH + FCS16_LENGTH
#define FLAG_LENGTH		3

#define OEM_ADDR		FLAG_ADDR + FLAG_LENGTH + FCS16_LENGTH
#define OEM_LENGTH		48

#define WMBUS_ADDR		OEM_ADDR + OEM_LENGTH + FCS16_LENGTH
#define WMBUS_LENGTH	16

#define PWHASH_ADDR		WMBUS_ADDR + WMBUS_LENGTH + FCS16_LENGTH
#define PWHASH_LENGTH	32

/***************************************************************************************/

#include <assert.h>
#include "fcs16.h"
#include "hashcalculator.h"
#include "eeprom.h"

cSysFile::cSysFile(const char *name, mode_t mode)
{
	fd = open(name, mode);
	if (fd < 0)
		printf("%s: can't open %s due to %s\n", __FUNCTION__, name, strerror(errno));
}

cSysFile::~cSysFile()
{
	if (fd)
		close(fd);
	fd = -1;
}

size_t cSysFile::rd(char *buf, const size_t len, const off_t offset)
{
	if ((fd > 0) && (buf))
	{
		if (offset > 0)
		{
			if (SIZEOF_I2C_EEPROM <= offset)
				return -1;
			lseek(fd, offset, SEEK_SET);
		}
		return read(fd, buf, len);
	}
	return -1;
}

size_t cSysFile::wr(const char *buf, const size_t len, const off_t offset)
{
	if ((fd > 0) && (buf))
	{
		if (offset > 0)
		{
			if (SIZEOF_I2C_EEPROM <= offset)
				return -1;
			lseek(fd, offset, SEEK_SET);
		}
		return write(fd, buf, len);
	}
	return -1;
}

void cEEPROM::print_hex(const char *data, const int len)
{
	int count = 0;
	while (count < len)
	{
		char Text[100], tmpText[10];
		int i;
		int k = ((len - count) > 16) ? 16 : (len - count);

		Text[0] = 0x00;
		for (i = 0; i < k; i++)
		{
			snprintf(tmpText, sizeof(tmpText), "%02X ", data[count + i]);
			strncat(Text, tmpText, 3);
		}
		while (i++ < 16)
			strncat(Text, "   ", 3);

		strncat(Text, "| ", 2);
		for (i = 0; i < k; i++)
		{
			if ((data[count + i] >= 32) && (data[count + i] <= 127))
				strncat(Text, &data[count + i], 1);
			else
				strncat(Text, " ", 1);
		}
		printf("  %s\n", Text);
		count += 16;
	}
}

int cEEPROM::erase()
{
	cSysFile fd(DEVICE_I2C_EEPROM, O_WRONLY);
	char buf[SIZEOF_I2C_EEPROM];

	for (int i = 0; i < SIZEOF_I2C_EEPROM; i++)
		buf[i] = 0x00;

	int len = fd.wr(buf, SIZEOF_I2C_EEPROM, 0);
	if (len != SIZEOF_I2C_EEPROM)
		printf("%s: Warning! wrote %d of %d\n", __FUNCTION__, len, SIZEOF_I2C_EEPROM);

	return EC_OK;
}

int cEEPROM::hexdump()
{
	cSysFile fd(DEVICE_I2C_EEPROM, O_RDONLY);
	char buf[SIZEOF_I2C_EEPROM];

	int len = fd.rd(buf, SIZEOF_I2C_EEPROM, 0);
	if (len != SIZEOF_I2C_EEPROM)
		printf("%s: Warning! read %d of %d\n", __FUNCTION__, len, SIZEOF_I2C_EEPROM);
	print_hex(buf, len);

	return EC_OK;
}

const char *cEEPROM::ec2str(const int ec)
{
	switch (ec)
	{
		default:
		case EC_UNSPECIFIED:
			return (const char *) "unspecified error";
		case EC_OK:
			return (const char *) "no error";
		case EC_FCS16:
			return (const char *) "fcs16 error";
		case EC_FORMAT:
			return (const char *) "format error";
		case EC_HASH:
			return (const char *) "hash error";
	}
}

#define UINT8_SIZE		1

int cEEPROM::setUint(const off_t offset, const uint8_t value)
{
	cSysFile fd(DEVICE_I2C_EEPROM, O_WRONLY);
	char buf[UINT8_SIZE + FCS16_LENGTH];

	buf[0] = value & 0x0ff;
	unsigned short fcs = calc_fcs16((unsigned char*) buf, UINT8_SIZE);
	buf[UINT8_SIZE + 1] = (fcs >> 8) & 0x0ff;
	buf[UINT8_SIZE + 0] = (fcs) & 0x0ff;

	int len = fd.wr(buf, UINT8_SIZE + FCS16_LENGTH, offset);
	if (len != UINT8_SIZE + FCS16_LENGTH)
	{
		printf("%s: Warning! wrote %d of %d\n", __FUNCTION__, len, UINT8_SIZE + FCS16_LENGTH);
		return EC_UNSPECIFIED;
	}
	return EC_OK;
}

int cEEPROM::getUint(const off_t offset, uint8_t &value)
{
	cSysFile fd(DEVICE_I2C_EEPROM, O_RDONLY);
	char buf[UINT8_SIZE + FCS16_LENGTH];

	int len = fd.rd(buf, UINT8_SIZE + FCS16_LENGTH, offset);
	if (len == UINT8_SIZE + FCS16_LENGTH)
	{
		if (true == verify_fcs16((unsigned char*) buf, UINT8_SIZE + FCS16_LENGTH))
		{
			value = buf[0];
			return EC_OK;
		}
		// printf("%s: fcs16 error\n", __FUNCTION__);
		return EC_FCS16;
	}
	// printf("%s: Warning! read %d of %d\n", __FUNCTION__, len, UINT8_SIZE + FCS16_LENGTH);
	return EC_UNSPECIFIED;
}

#define UINT16_SIZE		2

int cEEPROM::setUint(const off_t offset, const uint16_t value)
{
	cSysFile fd(DEVICE_I2C_EEPROM, O_WRONLY);
	char buf[UINT16_SIZE + FCS16_LENGTH];

	buf[0] = value & 0x0ff;
	buf[1] = (value >> 8) & 0x0ff;
	unsigned short fcs = calc_fcs16((unsigned char*) buf, UINT16_SIZE);
	buf[UINT16_SIZE + 1] = (fcs >> 8) & 0x0ff;
	buf[UINT16_SIZE + 0] = (fcs) & 0x0ff;

	int len = fd.wr(buf, UINT16_SIZE + FCS16_LENGTH, offset);
	if (len != UINT16_SIZE + FCS16_LENGTH)
	{
		printf("%s: Warning! wrote %d of %d\n", __FUNCTION__, len, UINT16_SIZE + FCS16_LENGTH);
		return EC_UNSPECIFIED;
	}
	return EC_OK;
}

#define UINT32_SIZE		4

int cEEPROM::setUint(const off_t offset, const uint32_t value)
{
	cSysFile fd(DEVICE_I2C_EEPROM, O_WRONLY);
	char buf[UINT32_SIZE + FCS16_LENGTH];

	buf[0] = value & 0x0ff;
	buf[1] = (value >> 8) & 0x0ff;
	buf[2] = (value >> 16) & 0x0ff;
	buf[3] = (value >> 24) & 0x0ff;
	unsigned short fcs = calc_fcs16((unsigned char*) buf, UINT32_SIZE);
	buf[UINT32_SIZE + 1] = (fcs >> 8) & 0x0ff;
	buf[UINT32_SIZE + 0] = (fcs) & 0x0ff;

	int len = fd.wr(buf, UINT32_SIZE + FCS16_LENGTH, offset);
	if (len != UINT32_SIZE + FCS16_LENGTH)
	{
		printf("%s: Warning! wrote %d of %d\n", __FUNCTION__, len, UINT32_SIZE + FCS16_LENGTH);
		return EC_UNSPECIFIED;
	}
	return EC_OK;
}

int cEEPROM::getUint(const off_t offset, uint16_t &value)
{
	cSysFile fd(DEVICE_I2C_EEPROM, O_RDONLY);
	char buf[UINT16_SIZE + FCS16_LENGTH];

	int len = fd.rd(buf, UINT16_SIZE + FCS16_LENGTH, offset);
	if (len == UINT16_SIZE + FCS16_LENGTH)
	{
		if (true == verify_fcs16((unsigned char*) buf, UINT16_SIZE + FCS16_LENGTH))
		{
			value = buf[0] | (buf[1] << 8);
			return EC_OK;
		}
		// printf("%s: fcs16 error\n", __FUNCTION__);
		return EC_FCS16;
	}
	// printf("%s: Warning! read %d of %d\n", __FUNCTION__, len, UINT16_SIZE + FCS16_LENGTH);
	return EC_UNSPECIFIED;
}

int cEEPROM::getUint(const off_t offset, uint32_t &value)
{
	cSysFile fd(DEVICE_I2C_EEPROM, O_RDONLY);
	char buf[UINT32_SIZE + FCS16_LENGTH];

	int len = fd.rd(buf, UINT32_SIZE + FCS16_LENGTH, offset);
	if (len == UINT32_SIZE + FCS16_LENGTH)
	{
		if (true == verify_fcs16((unsigned char*) buf, UINT32_SIZE + FCS16_LENGTH))
		{
			value = buf[0] | (buf[1] << 8) | (buf[2] << 16) | (buf[3] << 24);
			return EC_OK;
		}
		// printf("%s: fcs16 error\n", __FUNCTION__);
		return EC_FCS16;
	}
	// printf("%s: Warning! read %d of %d\n", __FUNCTION__, len, UINT16_SIZE + FCS16_LENGTH);
	return EC_UNSPECIFIED;
}

int cEEPROM::setData(const off_t offset, const char *value, const size_t length)
{
	if (value && (SIZEOF_I2C_EEPROM >= (length + FCS16_LENGTH)))
	{
		cSysFile fd(DEVICE_I2C_EEPROM, O_WRONLY);
		char buf[SIZEOF_I2C_EEPROM];
		memcpy(buf, value, length);
		unsigned short fcs = calc_fcs16((unsigned char*) buf, length);
		buf[length + 1] = (fcs >> 8) & 0x0ff;
		buf[length + 0] = (fcs) & 0x0ff;

		int len = fd.wr(buf, length + FCS16_LENGTH, offset);
		if (len == (int) (length + FCS16_LENGTH))
			return EC_OK;

		printf("%s: Warning! wrote %d of %d\n", __FUNCTION__, len, length + FCS16_LENGTH);
	}
	return EC_UNSPECIFIED;
}

int cEEPROM::getData(const off_t offset, char *value, const size_t length)
{
	if (value && (SIZEOF_I2C_EEPROM >= (length + FCS16_LENGTH)))
	{
		cSysFile fd(DEVICE_I2C_EEPROM, O_RDONLY);
		char buf[SIZEOF_I2C_EEPROM];

		int len = fd.rd(buf, length + FCS16_LENGTH, offset);
		if (len == (int) (length + FCS16_LENGTH))
		{
			if (true == verify_fcs16((unsigned char*) buf, length + FCS16_LENGTH))
			{
				memcpy(value, buf, length);
				return EC_OK;
			}
			// printf("%s: fcs16 error\n", __FUNCTION__);
			return EC_FCS16;
		}
		// printf("%s: Warning! read %d of %d\n", __FUNCTION__, len, length + FCS16_LENGTH);
	}
	return EC_UNSPECIFIED;
}

int cEEPROM::setVersion(const uint8_t value)
{
	return setUint(VERSION_ADDR, value);
}

int cEEPROM::getVersion(uint8_t &value)
{
	return getUint(VERSION_ADDR, value);
}

int cEEPROM::setBc(const uint32_t value)
{
	return setUint(BC_ADDR, value);
}

int cEEPROM::getBc(uint32_t &value)
{
	return getUint(BC_ADDR, value);
}

int cEEPROM::setOsc(const uint32_t value)
{
	return setUint(OSC_ADDR, value);
}

int cEEPROM::getOsc(uint32_t &value)
{
	return getUint(OSC_ADDR, value);
}

int cEEPROM::setHwid(const char *value, const size_t length)
{
	char buf[HWID_LENGTH];
	memset(buf, 0, HWID_LENGTH);
	memcpy(buf, value, min(HWID_LENGTH, length));
	return setData(HWID_ADDR, buf, HWID_LENGTH);
}

int cEEPROM::getHwid(char *value, size_t &length)
{
	length = HWID_LENGTH;
	return getData(HWID_ADDR, value, length);
}

int cEEPROM::setSwid(const char *value, const size_t length)
{
	char buf[SWID_LENGTH];
	memset(buf, 0, SWID_LENGTH);
	memcpy(buf, value, min(SWID_LENGTH, length));
	return setData(SWID_ADDR, buf, SWID_LENGTH);
}

int cEEPROM::getSwid(char *value, size_t &length)
{
	length = SWID_LENGTH;
	return getData(SWID_ADDR, value, length);
}

int cEEPROM::setSerial(const char *value, const size_t length)
{
	char buf[SN_LENGTH];
	memset(buf, 0, SN_LENGTH);
	memcpy(buf, value, min(SN_LENGTH, length));
	return setData(SN_ADDR, buf, SN_LENGTH);
}

int cEEPROM::getSerial(char *value, size_t &length)
{
	length = SN_LENGTH;
	return getData(SN_ADDR, value, length);
}

int cEEPROM::setPdate(const char *value, const size_t length)
{
	char buf[PDATE_LENGTH];
	memset(buf, 0, PDATE_LENGTH);
	memcpy(buf, value, min(PDATE_LENGTH, length));
	return setData(PDATE_ADDR, buf, PDATE_LENGTH);
}

int cEEPROM::getPdate(char *value, size_t &length)
{
	length = PDATE_LENGTH;
	return getData(PDATE_ADDR, value, length);
}

int cEEPROM::setHwver(const char *value, const size_t length)
{
	char buf[HWVER_LENGTH];
	memset(buf, 0, HWVER_LENGTH);
	memcpy(buf, value, min(HWVER_LENGTH, length));
	return setData(HWVER_ADDR, buf, HWVER_LENGTH);
}

int cEEPROM::getHwver(char *value, size_t &length)
{
	length = HWVER_LENGTH;
	return getData(HWVER_ADDR, value, length);
}

int cEEPROM::setMacEth0(const char *value, const size_t length)
{
	char buf[MAC_LENGTH];
	memset(buf, 0, MAC_LENGTH);
	memcpy(buf, value, min(MAC_LENGTH, length));

	for (int i = 0; i < MAC_LENGTH; i++)
	{
		if (((i == 2) || (i == 5) || (i == 8) || (i == 11) || (i == 14)) && (buf[i] == ':'))
			continue;
		else if ((buf[i] >= 'A') && (buf[i] <= 'F'))
			continue;
		else if ((buf[i] >= 'a') && (buf[i] <= 'f'))
			continue;
		else if ((buf[i] >= '0') && (buf[i] <= '9'))
			continue;
		else
			return EC_FORMAT;
	}
	return setData(MAC0_ADDR, buf, MAC_LENGTH);
}

int cEEPROM::getMacEth0(char *value, size_t &length)
{
	length = MAC_LENGTH;
	return getData(MAC0_ADDR, value, length);
}

int cEEPROM::setMacEth1(const char *value, const size_t length)
{
	char buf[MAC_LENGTH];
	memset(buf, 0, MAC_LENGTH);
	memcpy(buf, value, min(MAC_LENGTH, length));

	for (int i = 0; i < MAC_LENGTH; i++)
	{
		if (((i == 2) || (i == 5) || (i == 8) || (i == 11) || (i == 14)) && (buf[i] == ':'))
			continue;
		else if ((buf[i] >= 'A') && (buf[i] <= 'F'))
			continue;
		else if ((buf[i] >= 'a') && (buf[i] <= 'f'))
			continue;
		else if ((buf[i] >= '0') && (buf[i] <= '9'))
			continue;
		else
			return EC_FORMAT;
	}
	return setData(MAC1_ADDR, buf, MAC_LENGTH);
}

int cEEPROM::getMacEth1(char *value, size_t &length)
{
	length = MAC_LENGTH;
	return getData(MAC1_ADDR, value, length);
}

int cEEPROM::setKey(const char *value, const size_t length)
{
	char buf[KEY_LENGTH];
	memset(buf, 0, KEY_LENGTH);
	memcpy(buf, value, min(KEY_LENGTH, length));
	return setData(KEY_ADDR, buf, KEY_LENGTH);
}

int cEEPROM::getKey(char *value, size_t &length)
{
	length = KEY_LENGTH;
	return getData(KEY_ADDR, value, length);
}

int cEEPROM::setFlag(const char *value, const size_t length)
{
	char buf[FLAG_LENGTH];
	memset(buf, 0, FLAG_LENGTH);
	memcpy(buf, value, min(FLAG_LENGTH, length));
	return setData(FLAG_ADDR, buf, FLAG_LENGTH);
}

int cEEPROM::getFlag(char *value, size_t &length)
{
	length = FLAG_LENGTH;
	return getData(FLAG_ADDR, value, length);
}

int cEEPROM::setOem(const char *value, const size_t length)
{
	char buf[OEM_LENGTH];
	memset(buf, 0, OEM_LENGTH);
	memcpy(buf, value, min(OEM_LENGTH, length));
	return setData(OEM_ADDR, buf, OEM_LENGTH);
}

int cEEPROM::getOem(char *value, size_t &length)
{
	length = OEM_LENGTH;
	return getData(OEM_ADDR, value, length);
}

int cEEPROM::setWmbus(const char *value, const size_t length)
{
	char buf[WMBUS_LENGTH];
	memset(buf, 0, WMBUS_LENGTH);
	memcpy(buf, value, min(WMBUS_LENGTH, length));

	for (int i = 0; i < WMBUS_LENGTH; i++)
	{
		if ((buf[i] >= 'A') && (buf[i] <= 'F'))
			continue;
		else if ((buf[i] >= 'a') && (buf[i] <= 'f'))
			continue;
		else if ((buf[i] >= '0') && (buf[i] <= '9'))
			continue;
		else
			return EC_FORMAT;
	}
	return setData(WMBUS_ADDR, buf, WMBUS_LENGTH);
}

int cEEPROM::getWmbus(char *value, size_t &length)
{
	length = WMBUS_LENGTH;
	return getData(WMBUS_ADDR, value, length);
}

int cEEPROM::setPwHash(const char *value, const size_t length)
{
	return setData(PWHASH_ADDR, value, PWHASH_LENGTH);
}

int cEEPROM::getPwHash(char *value, size_t &length)
{
	length = PWHASH_LENGTH;
	return getData(PWHASH_ADDR, value, length);
}

void cEEPROM::createHash(char* acPwdHash, const char* pDataToHash, const size_t length)
{
	SHA256Calculator hashCalc;
	hashCalc.AddData(pDataToHash, length);
	string strTemp;
	strTemp.clear();
	hashCalc.GetHashResult(strTemp);

	// printf("%s: %s\n", __FUNCTION__, strTemp.c_str());

	char acTemp[3] =
	{ 0, 0, 0 };

	unsigned int uiTemp;
	unsigned int j = 0;
	for (unsigned int i = 0; i < strTemp.length(); i += 2, j++)
	{
		acTemp[0] = strTemp[i];
		acTemp[1] = strTemp[i + 1];
		acTemp[2] = 0;
		sscanf(acTemp, "%2x", &uiTemp);
		acPwdHash[j] = (char) uiTemp;
	}
}

int cEEPROM::setPassword(const char *value, const size_t length)
{
	char buf[PWHASH_LENGTH];
	memset(buf, 0, PWHASH_LENGTH);
	createHash(buf, value, length);

	return setPwHash(buf, PWHASH_LENGTH);
}

int cEEPROM::verifyPassword(const char *value, const size_t length)
{
	char buf[PWHASH_LENGTH];
	memset(buf, 0, PWHASH_LENGTH);
	createHash(buf, value, length);

	char tmp[PWHASH_LENGTH];
	int rc = EC_OK;
	size_t len;
	if (EC_OK == (rc = getPwHash(tmp, len)))
	{
		for (int i = 0; i < PWHASH_LENGTH; i++)
		{
			if (buf[i] != tmp[i])
				return EC_HASH;
		}
	}
	return rc;
}
