/*****************************************************************
 *
 * EEPROM - Demo Application
 *
 * Copyright (c) 2011 Birger Bauch, Dr. Neuhaus Telekommunikation GmbH
 *
 *****************************************************************/

#include "eeprom.h"

#define VERSION "v0.01 - "__DATE__" "__TIME__

/****************************************************************/

static int print_help(void)
{
	printf("\nI2C EEPROM - Demo Application\n");
	printf("\n");
	printf("\t-set <version> <bc> <osc> <hwid> <swid> <serial> <pdate> <hwver> <mac0> <mac1> <key> <flag> <oem> <wmbus> <password>...\n");
	printf("\n");
	printf("\t-erase                   | erase eeprom content\n");
	printf("\t-hexdump                 | dump eeprom content\n");
	printf("\t-get                     | read parameter list\n");
	printf("\t-rd <parameter>          | read parameter\n");
	printf("\t-rx <parameter>          | read parameter, display as hex\n");
	printf("\t-wr <parameter> <value>  | write parameter\n");
	printf("\t-verify <value>          | verify password\n");
	printf("\n");
	printf("\t-v version\n");
	printf("\t-? help\n");
	printf("\n");

	return EXIT_SUCCESS;
}

static int print_error(const char *Text)
{
	printf("error: %s\n", Text);
	return EXIT_SUCCESS;
}

static char* get_mobile_copyright()
{
	char *cText = NULL;
#	define COPYRIGHT_SIZE	1024
	if (NULL != (cText = (char *) malloc(COPYRIGHT_SIZE)))
	{
		char tmpText[120];

		strcpy(cText, "\nEEPROM Demo Application");

#ifdef __GNUC__
		snprintf(tmpText, sizeof(tmpText), "\n%s - GCC %s", VERSION, __VERSION__);
#elif defined _MSC_VER
		snprintf(tmpText, sizeof(tmpText), "\n%s - MSC %d", VERSION, _MSC_VER);
#else
		snprintf(tmpText, sizeof(tmpText), "\n%s", VERSION);
#endif
		if ((strlen(cText) + strlen(tmpText) + 2) < COPYRIGHT_SIZE)
			strcat(cText, tmpText);
		strcpy(tmpText, "\nCopyright (c) 2011 Dr.Neuhaus Telekommunikation GmbH\n");
		if ((strlen(cText) + strlen(tmpText) + 2) < COPYRIGHT_SIZE)
			strcat(cText, tmpText);
		strcpy(tmpText, "\n");
		if ((strlen(cText) + strlen(tmpText) + 2) < COPYRIGHT_SIZE)
			strcat(cText, tmpText);
	}
	return cText;
}

static void print_value(cEEPROM &eeprom, char *buf, const int len, const int display_hex)
{
	if (display_hex)
		eeprom.print_hex(buf, len);
	else
	{
		buf[len] = 0x00;
		printf("%s\n", buf);
	}
}

int main(int argc, char **argv)
{
	int arg_index = 1;
	int display_hex = false;
	char eeprom_parameter[255] = "";
	char eeprom_value[255] = "";

	enum
	{
		UNDEFINED, EEPROM_ERASE, EEPROM_HEXDUMP, EEPROM_READ, EEPROM_WRITE, EEPROM_GET_ALL, EEPROM_SET_ALL, EEPROM_VERIFY
	} action = UNDEFINED;

	while (arg_index < argc)
	{
		if (0 == strcmp(argv[arg_index], "-v"))
		{
			char *cText = get_mobile_copyright();
			if (cText != NULL)
			{
				printf("%s", cText);
				free(cText);
			}
			return EXIT_SUCCESS;
		}
		else if (0 == strcmp(argv[arg_index], "-?"))
			return print_help();
		else if (0 == strcmp(argv[arg_index], "-h"))
			return print_help();
		else if (0 == strcmp(argv[arg_index], "-erase"))
		{
			action = EEPROM_ERASE;
			break;
		}
		else if (0 == strcmp(argv[arg_index], "-hexdump"))
		{
			action = EEPROM_HEXDUMP;
			break;
		}
		else if (0 == strcmp(argv[arg_index], "-get"))
		{
			action = EEPROM_GET_ALL;
			break;
		}
		else if (0 == strcmp(argv[arg_index], "-set"))
		{
			action = EEPROM_SET_ALL;
			break;
		}
		else if (0 == strcmp(argv[arg_index], "-rd"))
		{
			arg_index++;
			if (arg_index < argc)
			{
				strncpy(eeprom_parameter, argv[arg_index], sizeof(eeprom_parameter));
				action = EEPROM_READ;
			}
			break;
		}
		else if (0 == strcmp(argv[arg_index], "-rx"))
		{
			arg_index++;
			if (arg_index < argc)
			{
				strncpy(eeprom_parameter, argv[arg_index], sizeof(eeprom_parameter));
				action = EEPROM_READ;
				display_hex = true;
			}
			break;
		}
		else if (0 == strcmp(argv[arg_index], "-wr"))
		{
			arg_index++;
			if (arg_index < argc)
			{
				strncpy(eeprom_parameter, argv[arg_index], sizeof(eeprom_parameter));
				arg_index++;
				if (arg_index < argc)
				{
					strncpy(eeprom_value, argv[arg_index], sizeof(eeprom_value));
					action = EEPROM_WRITE;
				}
			}
			break;
		}
		else if (0 == strcmp(argv[arg_index], "-verify"))
		{
			arg_index++;
			if (arg_index < argc)
			{
				strncpy(eeprom_parameter, argv[arg_index], sizeof(eeprom_parameter));
				action = EEPROM_VERIFY;
			}
			break;
		}
		else
			return print_error("unknown parameter, try -? for help");

		arg_index++;
	}

	cEEPROM eeprom;
	int retval = EC_OK;
	uint8_t i8 = 0;
	uint32_t i32 = 0;
	size_t buf_len = 0;
	char buf[512];
	memset(buf, 0, sizeof(buf));

	switch (action)
	{
		default:
			return print_error("missing parameter, try -? for help");

		case EEPROM_ERASE:
			retval = eeprom.erase();
			break;

		case EEPROM_HEXDUMP:
			retval = eeprom.hexdump();
			break;

		case EEPROM_GET_ALL:
			if (0 == (retval = eeprom.getVersion(i8)))
				printf("version: %d\n", i8);
			else
				printf("version: %s\n", eeprom.ec2str(retval));

			if (0 == (retval = eeprom.getBc(i32)))
				printf("bc: %d\n", i32);
			else
				printf("bc: %s\n", eeprom.ec2str(retval));

			if (0 == (retval = eeprom.getOsc(i32)))
				printf("osc: %d\n", i32);
			else
				printf("osc: %s\n", eeprom.ec2str(retval));

			if (0 == (retval = eeprom.getHwid(buf, buf_len)))
			{
				buf[buf_len] = 0x00;
				printf("hwid: %s\n", buf);
			}
			else
				printf("hwid: %s\n", eeprom.ec2str(retval));

			if (0 == (retval = eeprom.getSwid(buf, buf_len)))
			{
				buf[buf_len] = 0x00;
				printf("swid: %s\n", buf);
			}
			else
				printf("swid: %s\n", eeprom.ec2str(retval));

			if (0 == (retval = eeprom.getSerial(buf, buf_len)))
			{
				buf[buf_len] = 0x00;
				printf("serial: %s\n", buf);
			}
			else
				printf("serial: %s\n", eeprom.ec2str(retval));

			if (0 == (retval = eeprom.getPdate(buf, buf_len)))
			{
				buf[buf_len] = 0x00;
				printf("pdate: %s\n", buf);
			}
			else
				printf("pdate: %s\n", eeprom.ec2str(retval));

			if (0 == (retval = eeprom.getHwver(buf, buf_len)))
			{
				buf[buf_len] = 0x00;
				printf("hwver: %s\n", buf);
			}
			else
				printf("hwver: %s\n", eeprom.ec2str(retval));

			if (0 == (retval = eeprom.getMacEth0(buf, buf_len)))
			{
				buf[buf_len] = 0x00;
				printf("mac0: %s\n", buf);
			}
			else
				printf("mac0: %s\n", eeprom.ec2str(retval));

			if (0 == (retval = eeprom.getMacEth1(buf, buf_len)))
			{
				buf[buf_len] = 0x00;
				printf("mac1: %s\n", buf);
			}
			else
				printf("mac1: %s\n", eeprom.ec2str(retval));

			if (0 == (retval = eeprom.getKey(buf, buf_len)))
			{
				buf[buf_len] = 0x00;
				printf("key: %s\n", buf);
			}
			else
				printf("key: %s\n", eeprom.ec2str(retval));

			if (0 == (retval = eeprom.getFlag(buf, buf_len)))
			{
				buf[buf_len] = 0x00;
				printf("flag: %s\n", buf);
			}
			else
				printf("flag: %s\n", eeprom.ec2str(retval));

			if (0 == (retval = eeprom.getOem(buf, buf_len)))
			{
				buf[buf_len] = 0x00;
				printf("oem: %s\n", buf);
			}
			else
				printf("oem: %s\n", eeprom.ec2str(retval));

			if (0 == (retval = eeprom.getWmbus(buf, buf_len)))
			{
				buf[buf_len] = 0x00;
				printf("wmbus: %s\n", buf);
			}
			else
				printf("wmbus: %s\n", eeprom.ec2str(retval));

			return EXIT_SUCCESS;

		case EEPROM_SET_ALL:
			if ((arg_index < argc) && (retval == 0))
			{
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setVersion((uint8_t) atoi(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setBc((uint32_t) atoi(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setOsc((uint32_t) atoi(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setHwid(argv[arg_index], strlen(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setSwid(argv[arg_index], strlen(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setSerial(argv[arg_index], strlen(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setPdate(argv[arg_index], strlen(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setHwver(argv[arg_index], strlen(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setMacEth0(argv[arg_index], strlen(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setMacEth1(argv[arg_index], strlen(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setKey(argv[arg_index], strlen(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setFlag(argv[arg_index], strlen(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setOem(argv[arg_index], strlen(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setWmbus(argv[arg_index], strlen(argv[arg_index]));
				arg_index++;
				if ((arg_index < argc) && (retval == 0))
					retval = eeprom.setPassword(argv[arg_index], strlen(argv[arg_index]));
			}
			else
				return print_error("unknown parameter, try -? for help");
			break;

		case EEPROM_READ:
			if (0 == strcmp(eeprom_parameter, "version"))
			{
				uint8_t i = 0;
				if (0 == (retval = eeprom.getVersion(i)))
					printf("%d\n", i);
			}
			else if (0 == strcmp(eeprom_parameter, "bc"))
			{
				uint32_t i = 0;
				if (0 == (retval = eeprom.getBc(i)))
					printf("%d\n", i);
			}
			else if (0 == strcmp(eeprom_parameter, "osc"))
			{
				uint32_t i = 0;
				if (0 == (retval = eeprom.getOsc(i)))
					printf("%d\n", i);
			}
			else if (0 == strcmp(eeprom_parameter, "hwid"))
			{
				if (0 == (retval = eeprom.getHwid(buf, buf_len)))
					print_value(eeprom, buf, buf_len, display_hex);
			}
			else if (0 == strcmp(eeprom_parameter, "swid"))
			{
				if (0 == (retval = eeprom.getSwid(buf, buf_len)))
					print_value(eeprom, buf, buf_len, display_hex);
			}
			else if (0 == strcmp(eeprom_parameter, "serial"))
			{
				if (0 == (retval = eeprom.getSerial(buf, buf_len)))
					print_value(eeprom, buf, buf_len, display_hex);
			}
			else if (0 == strcmp(eeprom_parameter, "pdate"))
			{
				if (0 == (retval = eeprom.getPdate(buf, buf_len)))
					print_value(eeprom, buf, buf_len, display_hex);
			}
			else if (0 == strcmp(eeprom_parameter, "hwver"))
			{
				if (0 == (retval = eeprom.getHwver(buf, buf_len)))
					print_value(eeprom, buf, buf_len, display_hex);
			}
			else if (0 == strcmp(eeprom_parameter, "mac0"))
			{
				if (0 == (retval = eeprom.getMacEth0(buf, buf_len)))
					print_value(eeprom, buf, buf_len, display_hex);
			}
			else if (0 == strcmp(eeprom_parameter, "mac1"))
			{
				if (0 == (retval = eeprom.getMacEth1(buf, buf_len)))
					print_value(eeprom, buf, buf_len, display_hex);
			}
			else if (0 == strcmp(eeprom_parameter, "key"))
			{
				if (0 == (retval = eeprom.getKey(buf, buf_len)))
					print_value(eeprom, buf, buf_len, display_hex);
			}
			else if (0 == strcmp(eeprom_parameter, "flag"))
			{
				if (0 == (retval = eeprom.getFlag(buf, buf_len)))
					print_value(eeprom, buf, buf_len, display_hex);
			}
			else if (0 == strcmp(eeprom_parameter, "oem"))
			{
				if (0 == (retval = eeprom.getOem(buf, buf_len)))
					print_value(eeprom, buf, buf_len, display_hex);
			}
			else if (0 == strcmp(eeprom_parameter, "wmbus"))
			{
				if (0 == (retval = eeprom.getWmbus(buf, buf_len)))
					print_value(eeprom, buf, buf_len, display_hex);
			}
			else
				return print_error("unknown parameter, try -? for help");
			break;

		case EEPROM_WRITE:
			if (0 == strcmp(eeprom_parameter, "version"))
				retval = eeprom.setVersion((uint8_t) atoi(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "bc"))
				retval = eeprom.setBc((uint32_t) atoi(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "osc"))
				retval = eeprom.setOsc((uint32_t) atoi(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "hwid"))
				retval = eeprom.setHwid(eeprom_value, strlen(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "swid"))
				retval = eeprom.setSwid(eeprom_value, strlen(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "serial"))
				retval = eeprom.setSerial(eeprom_value, strlen(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "pdate"))
				retval = eeprom.setPdate(eeprom_value, strlen(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "hwver"))
				retval = eeprom.setHwver(eeprom_value, strlen(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "mac0"))
				retval = eeprom.setMacEth0(eeprom_value, strlen(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "mac1"))
				retval = eeprom.setMacEth1(eeprom_value, strlen(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "key"))
				retval = eeprom.setKey(eeprom_value, strlen(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "flag"))
				retval = eeprom.setFlag(eeprom_value, strlen(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "oem"))
				retval = eeprom.setOem(eeprom_value, strlen(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "wmbus"))
				retval = eeprom.setWmbus(eeprom_value, strlen(eeprom_value));
			else if (0 == strcmp(eeprom_parameter, "password"))
				retval = eeprom.setPassword(eeprom_value, strlen(eeprom_value));
			else
				return print_error("unknown parameter, try -? for help");
			break;

		case EEPROM_VERIFY:
			if (EC_OK == (retval = eeprom.verifyPassword(eeprom_parameter, strlen(eeprom_parameter))))
				printf("  Password verification SUCCESSFULL!\n");
			else
				printf("  Password verification FAILED!\n");
			return EXIT_SUCCESS;
	}

	if (retval)
		print_error(eeprom.ec2str(retval));

	return EXIT_SUCCESS;
}
