/*****************************************************************
 *
 * EEPROM - Generic Class Interface for Hash Calculation
 *
 * Copyright (c) 2008 SSI, Dr. Neuhaus Telekommunikation GmbH
 *
 *****************************************************************/

#include "hashcalculator.h"

static unsigned int K256[64] =
{
	0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
	0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3, 0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
	0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
	0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
	0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13, 0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
	0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
	0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
	0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,	0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
};

HashCalculator::HashCalculator()
{
	iBufferPos = 0;
	iBlockSize = 0;
	acHashResult = NULL;
	iHashResultLen = 0;
	acBlockBuffer = NULL;
	llMessageSize = 0;
}

void HashCalculator::AddFile(const char* FileName)
{
	FILE* InputFile;
	int BytesRead;
	char Bytes[iBlockSize];

	if ((InputFile = fopen(FileName, "rb")) != NULL)
	{
		do
		{
			BytesRead = fread((void*) Bytes, 1, iBlockSize, InputFile);
			AddData(Bytes, BytesRead);
		} while (!feof(InputFile));
	}
}

void HashCalculator::AddData(const char* Data, int Len)
{
	if (acBlockBuffer != NULL)
	{
		for (int i = 0; i < Len; i++)
		{
			acBlockBuffer[iBufferPos++] = Data[i];
			llMessageSize += 8;

			if (iBufferPos == iBlockSize)
			{
				iBufferPos = 0;
				ProcessBlock((unsigned int*) acBlockBuffer);
			}
		}
	}
}

void HashCalculator::GetHashResult(char* ResultTarget)
{
	char TempResult[10];

	Finish();

	ResultTarget[0] = 0;

	for (int i = 0; i < iHashResultLen; i++)
	{
		snprintf(TempResult, 10, "%.8x", acHashResult[i]);
		strncat(ResultTarget, TempResult, 10);
	}
	Init();
}

void HashCalculator::GetHashResult(string& ResultTarget)
{
	char TempResult[10];

	Finish();

	ResultTarget.clear();

	for (int i = 0; i < iHashResultLen; i++)
	{
		snprintf(TempResult, 10, "%.8x", acHashResult[i]);
		ResultTarget += TempResult;
	}
	Init();
}

SHA256Calculator::SHA256Calculator()
{
	iBufferPos = 0;
	iBlockSize = 64;
	acHashResult = new unsigned int[8];
	iHashResultLen = 8;
	acBlockBuffer = new char[64];
	llMessageSize = 0;
	acHashResult[0] = 0x6a09e667;
	acHashResult[1] = 0xbb67ae85;
	acHashResult[2] = 0x3c6ef372;
	acHashResult[3] = 0xa54ff53a;
	acHashResult[4] = 0x510e527f;
	acHashResult[5] = 0x9b05688c;
	acHashResult[6] = 0x1f83d9ab;
	acHashResult[7] = 0x5be0cd19;
}

SHA256Calculator::~SHA256Calculator()
{
	if (acHashResult != NULL)
		delete[] acHashResult;

	if (acBlockBuffer != NULL)
		delete[] acBlockBuffer;
}

unsigned int SHA256Calculator::ROTR(unsigned int X, int n)
{
	return (X >> n) | (X << (32 - n));
}

unsigned int SHA256Calculator::SIGMA0(unsigned int X)
{
	return ROTR(X, 2) ^ ROTR(X, 13) ^ ROTR(X, 22);
}

unsigned int SHA256Calculator::SIGMA1(unsigned int X)
{
	return ROTR(X, 6) ^ ROTR(X, 11) ^ ROTR(X, 25);
}

unsigned int SHA256Calculator::Sigma0(unsigned int X)
{
	return ROTR(X, 7) ^ ROTR(X, 18) ^ (X >> 3);
}

unsigned int SHA256Calculator::Sigma1(unsigned int X)
{
	return ROTR(X, 17) ^ ROTR(X, 19) ^ (X >> 10);
}

unsigned int SHA256Calculator::Ch(unsigned int X, unsigned int Y, unsigned int Z)
{
	return (X & Y) ^ (~X & Z);
}

unsigned int SHA256Calculator::Maj(unsigned int X, unsigned int Y, unsigned int Z)
{
	return (X & Y) ^ (X & Z) ^ (Y & Z);
}

void SHA256Calculator::Init()
{
	iBufferPos = 0;
	llMessageSize = 0;
	acHashResult[0] = 0x6a09e667;
	acHashResult[1] = 0xbb67ae85;
	acHashResult[2] = 0x3c6ef372;
	acHashResult[3] = 0xa54ff53a;
	acHashResult[4] = 0x510e527f;
	acHashResult[5] = 0x9b05688c;
	acHashResult[6] = 0x1f83d9ab;
	acHashResult[7] = 0x5be0cd19;
}

void SHA256Calculator::ProcessBlock(unsigned int* BlockData)
{
	for (int t = 0; t < 16; t++)
	{
		W[t] = (BlockData[t] & 0xFF000000) >> 24;
		W[t] |= (BlockData[t] & 0x00FF0000) >> 8;
		W[t] |= (BlockData[t] & 0x0000FF00) << 8;
		W[t] |= (BlockData[t] & 0x000000FF) << 24;
	}

	for (int t = 16; t < 64; t++)
		W[t] = Sigma1(W[t - 2]) + W[t - 7] + Sigma0(W[t - 15]) + W[t - 16];

	a = acHashResult[0];
	b = acHashResult[1];
	c = acHashResult[2];
	d = acHashResult[3];
	e = acHashResult[4];
	f = acHashResult[5];
	g = acHashResult[6];
	h = acHashResult[7];

	for (int t = 0; t < 64; t++)
	{
		T1 = h + SIGMA1(e) + Ch(e, f, g) + K256[t] + W[t];
		T2 = SIGMA0(a) + Maj(a, b, c);
		h = g;
		g = f;
		f = e;
		e = d + T1;
		d = c;
		c = b;
		b = a;
		a = T1 + T2;
	}

	acHashResult[0] += a;
	acHashResult[1] += b;
	acHashResult[2] += c;
	acHashResult[3] += d;
	acHashResult[4] += e;
	acHashResult[5] += f;
	acHashResult[6] += g;
	acHashResult[7] += h;
}

void SHA256Calculator::Finish()
{
	if (iBufferPos < 56)
	{
		acBlockBuffer[iBufferPos++] = 0x80;

		for (int i = iBufferPos; i < 64; i++)
			acBlockBuffer[i] = 0x00;

		acBlockBuffer[56] = (char) ((llMessageSize & 0xFF00000000000000LL) >> 56);
		acBlockBuffer[57] = (char) ((llMessageSize & 0x00FF000000000000LL) >> 48);
		acBlockBuffer[58] = (char) ((llMessageSize & 0x0000FF0000000000LL) >> 40);
		acBlockBuffer[59] = (char) ((llMessageSize & 0x000000FF00000000LL) >> 32);
		acBlockBuffer[60] = (char) ((llMessageSize & 0x00000000FF000000LL) >> 24);
		acBlockBuffer[61] = (char) ((llMessageSize & 0x0000000000FF0000LL) >> 16);
		acBlockBuffer[62] = (char) ((llMessageSize & 0x000000000000FF00LL) >> 8);
		acBlockBuffer[63] = (char) (llMessageSize & 0x00000000000000FFLL);

		ProcessBlock((unsigned int*) acBlockBuffer);
	}
	else
	{
		acBlockBuffer[iBufferPos++] = 0x80;

		for (int i = iBufferPos; i < 64; i++)
			acBlockBuffer[i] = 0x00;

		ProcessBlock((unsigned int*) acBlockBuffer);

		for (int i = 0; i < 64; i++)
			acBlockBuffer[i] = 0x00;

		acBlockBuffer[56] = (char) ((llMessageSize & 0xFF00000000000000LL) >> 56);
		acBlockBuffer[57] = (char) ((llMessageSize & 0x00FF000000000000LL) >> 48);
		acBlockBuffer[58] = (char) ((llMessageSize & 0x0000FF0000000000LL) >> 40);
		acBlockBuffer[59] = (char) ((llMessageSize & 0x000000FF00000000LL) >> 32);
		acBlockBuffer[60] = (char) ((llMessageSize & 0x00000000FF000000LL) >> 24);
		acBlockBuffer[61] = (char) ((llMessageSize & 0x0000000000FF0000LL) >> 16);
		acBlockBuffer[62] = (char) ((llMessageSize & 0x000000000000FF00LL) >> 8);
		acBlockBuffer[63] = (char) (llMessageSize & 0x00000000000000FFLL);

		ProcessBlock((unsigned int*) acBlockBuffer);
	}
}
