/*****************************************************************
 *
 * Watchdog - Demo Application
 *
 * Copyright (c) 2011 Birger Bauch, Dr. Neuhaus Telekommunikation GmbH
 *
 *****************************************************************/

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <termios.h>
#include <errno.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <syslog.h>

#ifdef MEMLIB_ENABLED
#include "memlib.h"
#endif

#include <assert.h>

#include <linux/watchdog.h>

/****************************************************************/

#ifndef VERSION
#define VERSION "v0.02 - "__DATE__" "__TIME__
#endif

#ifndef false
#define	false	0
#endif

#ifndef true
#define	true	1
#endif

#define DEVICE_WD "/dev/watchdog"

/****************************************************************/

int disable_watchdog()
{
	int fd = open(DEVICE_WD, O_WRONLY);
	if (fd < 0)
	{
		printf("%s: can't open %s due to %s\n", __FUNCTION__, DEVICE_WD, strerror(errno));
		return -1;
	}
	int flags = WDIOS_DISABLECARD;
	if (0 > ioctl(fd, WDIOC_SETOPTIONS, &flags))
		printf("%s: ioctl error (%s)\n", __FUNCTION__, strerror(errno));

	close(fd);
	return 0;
}

int enable_watchdog(const int value)
{
	int fd = open(DEVICE_WD, O_WRONLY);
	if (fd < 0)
	{
		printf("%s: can't open %s due to %s\n", __FUNCTION__, DEVICE_WD, strerror(errno));
		return -1;
	}
	int timeout = value;
	if (0 > ioctl(fd, WDIOC_SETTIMEOUT, &timeout))
		printf("%s: ioctl error (%s)\n", __FUNCTION__, strerror(errno));
	else
		printf("%s: %d sec timeout\n", __FUNCTION__, timeout);

	close(fd);
	return 0;
}

int reset_watchdog()
{
	int fd = open(DEVICE_WD, O_WRONLY);
	if (fd < 0)
	{
		printf("%s: can't open %s due to %s\n", __FUNCTION__, DEVICE_WD, strerror(errno));
		return -1;
	}
	if (0 > ioctl(fd, WDIOC_KEEPALIVE, NULL))
		printf("%s: ioctl error (%s)\n", __FUNCTION__, strerror(errno));

	close(fd);
	return 0;
}

int bootstatus_watchdog()
{
	int i, fd = open(DEVICE_WD, O_RDONLY);
	if (fd < 0)
	{
		printf("%s: can't open %s due to %s\n", __FUNCTION__, DEVICE_WD, strerror(errno));
		return -1;
	}
	if (0 > ioctl(fd, WDIOC_GETBOOTSTATUS, &i))
		printf("%s: ioctl error (%s)\n", __FUNCTION__, strerror(errno));
	else
		printf("%s: boot_reason %d\n", __FUNCTION__, i);

	close(fd);
	return i;
}

static int print_help(void)
{
	printf("\nWatchdog - Demo Application\n");
	printf("\n");
	printf("\t-b get boot_reason\n");
//	printf("\t-disable\n");
	printf("\t-enable <value>\n");
	printf("\t-reset\n");
	printf("\n");
	printf("\t-v version\n");
	printf("\t-? help\n");
	printf("\n");

	return EXIT_SUCCESS;
}

static int print_error(const char *Text)
{
	printf("error: %s\n", Text);
	return EXIT_SUCCESS;
}

static char* get_mobile_copyright()
{
	char *cText = NULL;
#	define COPYRIGHT_SIZE	1024
	if (NULL != (cText = (char *) malloc(COPYRIGHT_SIZE)))
	{
		char tmpText[120];

		strcpy(cText, "\nWatchdog Demo Application");

#ifdef __GNUC__
		snprintf(tmpText, sizeof(tmpText), "\n%s - GCC %s", VERSION, __VERSION__);
#elif defined _MSC_VER
		snprintf(tmpText, sizeof(tmpText), "\n%s - MSC %d", VERSION, _MSC_VER);
#else
		snprintf(tmpText, sizeof(tmpText), "\n%s", VERSION);
#endif
		if ((strlen(cText) + strlen(tmpText) + 2) < COPYRIGHT_SIZE)
			strcat(cText, tmpText);
		strcpy(tmpText, "\nCopyright (c) 2011 Dr.Neuhaus Telekommunikation GmbH\n");
		if ((strlen(cText) + strlen(tmpText) + 2) < COPYRIGHT_SIZE)
			strcat(cText, tmpText);
		strcpy(tmpText, "\n");
		if ((strlen(cText) + strlen(tmpText) + 2) < COPYRIGHT_SIZE)
			strcat(cText, tmpText);
	}
	return cText;
}

int main(int argc, char **argv)
{
	char tmpText[80] = "";
	int arg_index = 1;
	int value = 0;

	enum
	{
		UNDEFINED, DISABLE_WD, ENABLE_WD, RESET_WD, BOOTREASON_WD,
	} action = UNDEFINED;

	while (arg_index < argc)
	{
		if (0 == strcmp(argv[arg_index], "-v"))
		{
			char *cText = get_mobile_copyright();
			if (cText != NULL)
			{
				printf("%s", cText);
				free(cText);
			}
			return EXIT_SUCCESS;
		}
		else if (0 == strcmp(argv[arg_index], "-?"))
			return print_help();
		else if (0 == strcmp(argv[arg_index], "-h"))
			return print_help();
//		else if (0 == strcmp(argv[arg_index], "-disable"))
//		{
//			action = DISABLE_WD;
//			break;
//		}
		else if (0 == strcmp(argv[arg_index], "-enable"))
		{
			arg_index++;
			if (arg_index < argc)
			{
				value = atoi(argv[arg_index]);
				action = ENABLE_WD;
				break;
			}
		}
		else if (0 == strcmp(argv[arg_index], "-reset"))
		{
			action = RESET_WD;
			break;
		}
		else if (0 == strcmp(argv[arg_index], "-b"))
		{
			action = BOOTREASON_WD;
			break;
		}
		else
		{
			snprintf(tmpText, sizeof(tmpText), "wrong parameter: %s, try -? for help", argv[arg_index]);
			return print_error(tmpText);
		}
		arg_index++;
	}

	switch (action)
	{
		default:
			return print_error("missing parameter, try -? for help");
		case ENABLE_WD:
			enable_watchdog(value);
			break;
		case DISABLE_WD:
			disable_watchdog();
			break;
		case RESET_WD:
			reset_watchdog();
			break;
		case BOOTREASON_WD:
			bootstatus_watchdog();
			break;
	}
	return EXIT_SUCCESS;
}
