#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <termios.h>
#include <pthread.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>

#define VERSION "0.3"

enum {NONE, CHAR, BREAK, PARERR};

static struct
{
	int Val;
	int Sym;
}

BdTab[] =
{
	{300,   B300},
	{1200,  B1200},
	{2400,  B2400},
	{9600,  B9600},
	{19200, B19200},
	{38400, B38400},
	{57600, B57600},
	{115200,B115200},
	{0,     0}
};

int tty_open(const char *name, long baud)
{
	int rc;
	int TTY;
	int i;
	struct termios Tattr;

	for (i = 0; BdTab[i].Val != 0; i++)
	{
		if (BdTab[i].Val == baud)
			break;
	}
	if (BdTab[i].Val == 0)
		return -1;

	TTY = open(name, O_RDWR | O_NONBLOCK);
	if (TTY < 0)
		return -1;

	rc = tcgetattr(TTY, &Tattr);
	if (rc < 0)
		return -1;

	cfmakeraw(&Tattr);
	Tattr.c_iflag &= ~(IXON | IXOFF | IGNBRK | IGNPAR);
	Tattr.c_iflag |= PARMRK | INPCK;

	Tattr.c_cflag &= !(CRTSCTS | CSTOPB | CSIZE | PARODD | PARENB);
	Tattr.c_cflag |= CLOCAL | CREAD | CS8;

	cfsetospeed(&Tattr, BdTab[i].Sym);
	cfsetispeed(&Tattr, BdTab[i].Sym);

	Tattr.c_cc[VMIN] = 1;
	Tattr.c_cc[VTIME] = 1;

	rc = tcsetattr(TTY, TCSANOW, &Tattr);
	if (rc < 0)
		return -1;

	return TTY;
}

int nextch (unsigned char ch, unsigned char* result)
{
	static int idx = 0;
	switch (idx)
	{
		case 0:
			if (ch == 0xff)
			{
                idx++;
				return NONE;
			}
			else
			{
				*result = ch;
				return CHAR;
			}
			break;

		case 1:
			if (ch == 0x00)
			{
                idx++;
				return NONE;
			}
			else if (ch == 0xFF)
			{
                idx = 0;
				*result = ch;
				return CHAR;
			}
			else
			{
                idx = 0;
				*result = ch;
				return PARERR;
			}
			break;

		case 2:
			if (ch == 0x00)
			{
                idx = 0;
				*result = ch;
				return BREAK;
			}
            else
			{
                idx = 0;
				*result = ch;
				return PARERR;
			}
			break;

		default:
			idx = 0;
			break;
	}
	return NONE;
}

int main (int argc, char **argv)
{
	char *port = "/dev/ttyAPP3";
	int fd, rc, i;
	fd_set rfds;
	char buf[100];
	char* p;
	time_t wait = 0;
    time_t start = time(0);

	fd = tty_open(port, 19200);
	if (fd < 0)
	{
		printf("*** tty open error %d on %s\n", errno, port);
		exit(1);
	}

	if (argc > 1)
	{
		if (argv[1][0] == '-')
		{
			printf("mbustest V " VERSION "\n");
            return 0;
		}

		for (i = 0, p = argv[1]; i < (sizeof(buf) - 1) && *p; p++)
		{
			if (isxdigit(*p))
				buf[i++] = *p;
		}
		buf[i] = 0;

		for (p = buf; i >= 2; p += 2, i -= 2)
		{
			int ch;
			if (sscanf(p, "%2x", &ch) == 1)
                rc = write(fd, &ch, 1);
		}
	}

    // make stdin non-blocking
    //rc = fcntl(0, F_SETFL, fcntl(0, F_GETFL) | O_NONBLOCK);

	for (;;)
	{
		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);
		FD_SET(0, &rfds);
		struct timespec to = {1, 0};
		rc = pselect(fd + 1, &rfds, 0, 0, &to, 0);
		if (rc > 0)
		{
			if (FD_ISSET(fd, &rfds))
			{
				int len = read(fd, buf, sizeof(buf));
				if (len > 0)
				{
					for (i = 0; i < len; i++)
					{
						unsigned char result;
						switch (nextch(buf[i], &result))
						{
							case NONE:
								break;
							case CHAR:
								printf("%02x ", result);
								break;
							case PARERR:
								printf("<error>");
								break;
							case BREAK:
								printf("<break>");
								break;
							default:
								break;
						}
					}
					printf("\n");
				}
			}
			if (FD_ISSET(0, &rfds))
			{
				int len = read(0, buf, sizeof(buf));
				if (len > 0)
				{
					if (buf[0] == 'q' || buf[0] == 'Q')
						return 0;
				}
			}
		}
		if (wait > 0)
		{
			if (time(0) > start + wait)
				break;
		}
	}
	return 0;
}
