/*****************************************************************
 *
 * TTY - Demo Application: send string to tty and display received data
 *
 * wmbus:
 * -d /dev/ttyS1 -b B19200 -x 0030
 *
 * plc:
 * -d /dev/ttyS2 -p -b B57600 -x 02003c3c00
 *
 * rs485:
 * -d /dev/ttyS0 -b B19200 -m -c HALLO
 *
 * rs485 7E1:
 * -d /dev/ttyS0 -b B9600 -e -m -x 2f3f210d0a
 *
 * rs485 Acrel PZ72:
 * -d /dev/ttyS0 -m -b B9600 -x 0203000000018439
 * -d /dev/ttyS0 -m -b B9600  -z 2 -y 10 -x 0203000000018439
 *
 * mbus:
 * -d /dev/ttyS4 -x 000b43000203030100bc1a
 *
 * Copyright (c) 2012 Birger Bauch, Dr. Neuhaus Telekommunikation GmbH
 *
 *****************************************************************/

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <termios.h>
#include <errno.h>
#include <ctype.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <syslog.h>

#include <linux/serial.h>

#ifdef MEMLIB_ENABLED
#include "memlib.h"
#endif

#include <assert.h>

/****************************************************************/

#ifndef VERSION
#define VERSION "v0.01 - "__DATE__" "__TIME__
#endif

#ifndef false
#define	false	0
#endif

#ifndef true
#define	true	1
#endif

#ifndef max
#define max(a,b)  (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)  (((a) < (b)) ? (a) : (b))
#endif

#define IO_BUFFER_SIZE 2000

#define APL_OK                0
#define APL_ERR_ASN1         -1
#define APL_ERR_MEMORY       -2
#define APL_ERR_UNKNOWN_PDU  -3
#define APL_ERR_MISSING_PARA -4
#define APL_ERR_WRONG_PARA	 -5
#define APL_ERR_UNKNOWN_TTY  -6
#define APL_ERR_IGNORE_DATA  -7
#define APL_ERR_IO           -8
#define APL_ERR_LOCKED       -9
#define APL_ERR_BUSY         -10
#define APL_ERR_PPPD         -11

#ifndef TIOCGRS485
#define TIOCGRS485	0x542e
#endif

#ifndef TIOCSRS485
#define TIOCSRS485	0x542f
#endif

int enable_custom_485_delay = false;
unsigned int custom_485_delay = 0;

int enable_custom_485_aftersend_delay = false;
unsigned int custom_485_aftersend_delay = 0;

/****************************************************************/

static void mob_print_data(const unsigned char *data, const int len)
{
	char Text[100], tmpText[10], timestamp[20]; 
	int count, i, k;


    time_t ltime; /* calendar time */
    ltime=time(NULL); /* get current cal time */
   
    count = 0;
	while (count < len)
	{

        snprintf(timestamp, sizeof(timestamp), "%s",asctime( localtime(&ltime) ) );
		strncat(timestamp, " | " , 3);

		if ((len - count) > 16)
			k = 16;
		else
			k = len - count;

		Text[0] = 0x00;

		strncat(Text, timestamp, 30);
		for (i = 0; i < k; i++)
		{
			snprintf(tmpText, sizeof(tmpText), "%02X ", data[count + i]);
			strncat(Text, tmpText, 3);
		}
		while (i++ < 16)
			strncat(Text, "   ", 3);

		strncat(Text, "| ", 2);
		for (i = 0; i < k; i++)
		{
			if ((data[count + i] >= 32) && (data[count + i] <= 127))
				strncat(Text, (char*) &data[count + i], 1);
			else
				strncat(Text, " ", 1);
		}
		printf("  %s\n", Text);

		count += 16;
	}
}

typedef struct
{
	char buf[IO_BUFFER_SIZE];
	char *wptr, *rptr;
} T_TTY_IO;

#if 0
static int io_get_count(char *rptr, char *wptr, const int size)
{
	if (rptr == wptr)
	return 0;
	else if (rptr < wptr)
	return (wptr - rptr);
	else
	return (size - (rptr - wptr));
}
#endif

static int io_get_available(char *rptr, char *wptr, const int size)
{
	if (rptr == wptr)
		return (size - 1);
	else if (rptr < wptr)
		return (size - 1 - (wptr - rptr));
	else
		return (size - 1 - (size - (rptr - wptr)));
}

static int io_init(T_TTY_IO *dst)
{
	if (dst)
	{
		dst->rptr = dst->wptr = dst->buf;
		return APL_OK;
	}
	else
		return APL_ERR_MEMORY;
}

int io_store_data(T_TTY_IO *dst, const char *buf, const int len)
{
	int rc = io_get_available(dst->rptr, dst->wptr, IO_BUFFER_SIZE);
	if (rc < len)
		return APL_ERR_IO;

	rc = min(rc, len);
	while (rc--)
	{
		*dst->wptr++ = *buf++;
		if (dst->wptr == (dst->buf + IO_BUFFER_SIZE))
			dst->wptr = dst->buf;
	}
	return APL_OK;
}

static int io_get_data(T_TTY_IO *src, char *dst, const int len)
{
	if (dst)
	{
		int i = 0;
		while ((src->rptr != src->wptr) && (i < len))
		{
			*dst++ = *src->rptr++;
			if (src->rptr == (src->buf + IO_BUFFER_SIZE))
				src->rptr = src->buf;
			i++;
		}
		return i;
	}
	return 0;
}

//add user defined flag into struct serial_rs485, see /usr/include/serial.h for details
#define SER_RS485_ECHO		(1 << 3)

static void tty_set_RS485(const int fd, const u_int32_t bd, const int enable_rs485, const int enable_echo)
{
	struct serial_rs485 rs485;
	ioctl(fd, TIOCGRS485, &rs485);
	if (enable_rs485)
	{
		if ((rs485.flags & SER_RS485_ENABLED) != SER_RS485_ENABLED)
		{
			if (enable_custom_485_delay)
			{
				rs485.delay_rts_before_send = custom_485_delay;
				if (enable_custom_485_aftersend_delay)
				{
					rs485.delay_rts_before_send |= (custom_485_aftersend_delay<<16);
				}
			}
			else
			{
				switch (bd)
				{
					default:
						rs485.delay_rts_before_send = 0x00020000 + 10;
						break;
					case B4800:
						rs485.delay_rts_before_send = 0x00320000 + 30;
						break;
					case B2400:
						rs485.delay_rts_before_send = 0x00320000 + 50;
						break;
					case B1200:
						rs485.delay_rts_before_send = 0x00320000 + 90;
						break;
					case 600:
						rs485.delay_rts_before_send = 0x00320000 + 160;
						break;
					case B300:
						rs485.delay_rts_before_send = 0x00320000 + 300;
						break;
				}
			}
			if (enable_echo)
				rs485.flags |= SER_RS485_ECHO;

			rs485.flags |= SER_RS485_ENABLED;
			ioctl(fd, TIOCSRS485, &rs485);
		}
	}
	else
	{
		if ((rs485.flags & SER_RS485_ENABLED) == SER_RS485_ENABLED)
		{
			rs485.flags &= ~SER_RS485_ENABLED;
			ioctl(fd, TIOCSRS485, &rs485);
		}
	}
}

#if 0
static int tty_isTxBufferEmpty(const int fd)
{
	int i;
	ioctl(fd, TIOCSERGETLSR, &i );
	return ((i & TIOCSER_TEMT) == TIOCSER_TEMT) ? true:false;
}
#endif

static void tty_set_DTR(const int fd, const int flag)
{
	int i;
	ioctl(fd, TIOCMGET, &i);
	if (flag)
	{
		if ((i & TIOCM_DTR) != TIOCM_DTR)
		{
			i |= TIOCM_DTR;
			ioctl(fd, TIOCMSET, &i);
		}
	}
	else
	{
		if ((i & TIOCM_DTR) == TIOCM_DTR)
		{
			i &= ~TIOCM_DTR;
			ioctl(fd, TIOCMSET, &i);
		}
	}
}

static void tty_set_RTS(const int fd, const int flag)
{
	int i;
	ioctl(fd, TIOCMGET, &i);
	if (flag)
	{
		if ((i & TIOCM_RTS) != TIOCM_RTS)
		{
			i |= TIOCM_RTS;
			ioctl(fd, TIOCMSET, &i);
		}
	}
	else
	{
		if ((i & TIOCM_RTS) == TIOCM_RTS)
		{
			i &= ~TIOCM_RTS;
			ioctl(fd, TIOCMSET, &i);
		}
	}
}

static int tty_get_DCD(const int fd)
{
	int i;
	ioctl(fd, TIOCMGET, &i);
	return ((i & TIOCM_CD) == TIOCM_CD) ? 1 : 0;
}

static int tty_get_DSR(const int fd)
{
	int i;
	ioctl(fd, TIOCMGET, &i);
	return ((i & TIOCM_DSR) == TIOCM_DSR) ? 1 : 0;
}

static int tty_get_RI(const int fd)
{
	int i;
	ioctl(fd, TIOCMGET, &i);
	return ((i & TIOCM_RI) == TIOCM_RI) ? 1 : 0;
}

static int tty_get_DTR(const int fd)
{
	int i;
	ioctl(fd, TIOCMGET, &i);
	return ((i & TIOCM_DTR) == TIOCM_DTR) ? 1 : 0;
}

static int tty_get_RTS(const int fd)
{
	int i;
	ioctl(fd, TIOCMGET, &i);
	return ((i & TIOCM_RTS) == TIOCM_RTS) ? 1 : 0;
}

static int tty_get_CTS(const int fd)
{
	int i;
	ioctl(fd, TIOCMGET, &i);
	return ((i & TIOCM_CTS) == TIOCM_CTS) ? 1 : 0;
}

static void get_tty_modem_lines(const int fd)
{
	printf("%s: CTS[%d] RTS[%d] DSR[%d] CD[%d] DTR[%d] RI[%d]\n", __FUNCTION__, (tty_get_CTS(fd) > 0) ? 1 : 0, (tty_get_RTS(fd) > 0) ? 1 : 0, (tty_get_DSR(fd) > 0) ? 1 : 0,
			(tty_get_DCD(fd) > 0) ? 1 : 0, (tty_get_DTR(fd) > 0) ? 1 : 0, (tty_get_RI(fd) > 0) ? 1 : 0);
}

static char* get_mobile_copyright()
{
	char *cText = NULL;
#	define COPYRIGHT_SIZE	1024
	if (NULL != (cText = (char *) malloc(COPYRIGHT_SIZE)))
	{
		char tmpText[120];

		strcpy(cText, "\nTTY Demo Application");

#ifdef __GNUC__
		snprintf(tmpText, sizeof(tmpText), "\n%s - GCC %s", VERSION, __VERSION__);
#elif defined _MSC_VER
		snprintf(tmpText, sizeof(tmpText), "\n%s - MSC %d", VERSION, _MSC_VER);
#else
		snprintf(tmpText, sizeof(tmpText), "\n%s", VERSION);
#endif
		if ((strlen(cText) + strlen(tmpText) + 2) < COPYRIGHT_SIZE)
			strcat(cText, tmpText);
		strcpy(tmpText, "\nCopyright (c) 2014 Dr.Neuhaus Telekommunikation GmbH\n");
		if ((strlen(cText) + strlen(tmpText) + 2) < COPYRIGHT_SIZE)
			strcat(cText, tmpText);
		strcpy(tmpText, "\n");
		if ((strlen(cText) + strlen(tmpText) + 2) < COPYRIGHT_SIZE)
			strcat(cText, tmpText);
	}
	return cText;
}

static int demo_running = true;

static void signal_handler(const int sig)
{
	//	DEBUG_MSG("[%s] received %d\n", __FUNCTION__, sig);

	switch (sig)
	{
		default:
			break;
		case SIGTERM: //kill
			exit(0);
			break;
		case SIGINT: //strg-C
			demo_running = false;
			break;
	}
}

static int print_help(void)
{
	printf("\nTTY - Demo Application: send string to tty and display received data\n");
	printf("\n");
	printf("\t-c set command string (ascii)\n");
	printf("\t-x set command string (hex) (e.g. 0001AA0F of \"00 01 AA 0F\")\n");
	printf("\t-s delay command string (default: 100 ms)\n");
	printf("\t-d set device (e.g. '/dev/ttyS0')\n");
	printf("\t-b set tty baudrate (default: B115200)\n");
#ifdef TIMESEC
	printf("\t-t set receive timeout (default: 3 sec)\n");
#else
	printf("\t-t set receive timeout (default: 3 sec)\n");
	printf("\t-T set receive timeout (default: 3000 msec)\n");
#endif
	printf("\t-p toggle RTS before transmit, simulate T_REQ for ST7580\n");
	printf("\t-r display raw received data (default: hex)\n");
	printf("\t-m enable RS485 mode, if supported\n");
	printf("\t-n enable RS485 echo, if supported\n");
	printf("\t-e enable 7E1 (default 8N1)\n");
	printf("\t-f enable 8E1 (default 8N1)\n");
	printf("\t-l display modem lines\n");
	printf("\t-o enable rtscts flowcontrol\n");
	printf("\t-y custom RS485 lastchar delay (in 1/10ms)\n");
	printf("\t-z custom RS485 after send delay (in 1/10ms)\n");
	printf("\n");
	printf("\t-v version\n");
	printf("\t-? help\n");
	printf("\n");

	return EXIT_SUCCESS;
}

static int print_error(const char *Text)
{
	printf("error: %s\n", Text);
	return EXIT_SUCCESS;
}

static int tty_init(const int fd, const u_int32_t bd, const int rtscts, const int enable_7E1, const int enable_8E1)
{
	//http://www.faqs.org/docs/Linux-HOWTO/Serial-Programming-HOWTO.html
	//http://www.mikrocontroller.net/articles/Ports_benutzen_(GCC)

	int rc;
	struct termios tio;

	//get current attributes
	rc = tcgetattr(fd, &tio);
	if (rc < 0)
		return rc;

	//clear struct for new port settings
	bzero(&tio, sizeof(tio));

	if ((enable_7E1) && (enable_8E1))
		return -1;

	if (enable_7E1)
	{
		tio.c_cflag = CS7 | PARENB | CREAD | CLOCAL;
	}
	else if (enable_8E1)
	{
		tio.c_cflag = CS8 | PARENB | CREAD | CLOCAL;
	}
	else
	{
		//CS8		8n1 (8bit,no parity,1 stopbit)
		//CREAD		enable receiving characters
		//CLOCAL	local connection, no modem control
		//CRTSCTS	output hardware flow control
		if (rtscts)
			tio.c_cflag = CS8 | CREAD | CLOCAL | CRTSCTS;
		else
			tio.c_cflag = CS8 | CREAD | CLOCAL;
	}

	//IGNPAR  : ignore bytes with parity errors
	tio.c_iflag = IGNPAR;
	//Raw output.
	tio.c_oflag = 0;
	//input mode (non-canonical, no echo,...)
	tio.c_lflag = 0;

	//set input baudrate to zero, than it will be equal to output baudrate
	rc = cfsetispeed(&tio, B0);
	if (rc < 0)
		return rc;

	//set output baudrate
	rc = cfsetospeed(&tio, bd);
	if (rc < 0)
		return rc;

	//now clean the modem line and activate the settings for the port
	tcflush(fd, TCIFLUSH);

	//update attributes
	return tcsetattr(fd, TCSANOW, &tio);
}

static int demo_select(const int fd, T_TTY_IO *tty, const int display_raw)
{
	struct timeval timeout;
	fd_set rd_set, wr_set, ex_set;
	int rc;

	FD_ZERO(&rd_set);
	FD_ZERO(&wr_set);
	FD_ZERO(&ex_set);

	FD_SET(fd, &rd_set);

	timeout.tv_sec = 0;
	timeout.tv_usec = 100000; //100ms
	rc = select(FD_SETSIZE, &rd_set, &wr_set, &ex_set, &timeout);
	if (rc > 0)
	{
		char buf[IO_BUFFER_SIZE];

		if (FD_ISSET(fd, &rd_set))
		{
			rc = read(fd, buf, sizeof(buf));
			if (rc > 0)
			{
				if (display_raw)
					io_store_data(tty, buf, rc);
				else
					mob_print_data((unsigned char *) buf, rc);
			}
			else
			{
				// DEBUG_MSG("[%s] error reading tty => disconnect\n", __FUNCTION__);
				return -1;
			}
		}
	}
	return rc;
}

int main(int argc, char **argv)
{
	char tmpText[80] = "";
	int fd, arg_index = 1;
	char device[80] = "";
	char command[255] = "";
	int command_length = 0;
	int baudrate = B115200;
#ifdef TIMESEC
	unsigned int timeout = 3;
#else
	unsigned int timeout = 3000;
#endif
	int display_raw = false;
	int display_modem = false;
	int enable_rs485 = false;
	int enable_rs485_echo = false;
	int plc_with_t_req = false;
	int enable_7E1 = false;
	int enable_8E1 = false;
	int enable_rtscts = false;
	unsigned int command_delay = 100;

#ifdef TIMESEC
	time_t start_time = time((time_t *) 0);
#else
	struct timeval t1, t2;
	gettimeofday(&t1, NULL);
#endif

	signal(SIGTERM, signal_handler);
	signal(SIGINT, signal_handler);

	while (arg_index < argc)
	{
		if (0 == strcmp(argv[arg_index], "-v"))
		{
			char *cText = get_mobile_copyright();
			if (cText != NULL)
			{
				printf("%s", cText);
				free(cText);
			}
			return EXIT_SUCCESS;
		}
		else if (0 == strcmp(argv[arg_index], "-y"))
		{
			enable_custom_485_delay = true;
			arg_index++;
			if (arg_index < argc)
			{
				custom_485_delay = strtoul(argv[arg_index], NULL, 0);
			}
			else
			{
				return print_error("missing parameter, try -? for help");
			}
		}
		else if (0 == strcmp(argv[arg_index], "-z"))
		{
			enable_custom_485_aftersend_delay = true;
			arg_index++;
			if (arg_index < argc)
			{
				custom_485_aftersend_delay = strtoul(argv[arg_index], NULL, 0);
			}
			else
			{
				return print_error("missing parameter, try -? for help");
			}
		}
		else if (0 == strcmp(argv[arg_index], "-b"))
		{
			arg_index++;
			if (arg_index < argc)
			{
				if (0 == strcmp(argv[arg_index], "B460800"))
					baudrate = B460800;
				else if (0 == strcmp(argv[arg_index], "B230400"))
					baudrate = B230400;
				else if (0 == strcmp(argv[arg_index], "B115200"))
					baudrate = B115200;
				else if (0 == strcmp(argv[arg_index], "B57600"))
					baudrate = B57600;
				else if (0 == strcmp(argv[arg_index], "B38400"))
					baudrate = B38400;
				else if (0 == strcmp(argv[arg_index], "B19200"))
					baudrate = B19200;
				else if (0 == strcmp(argv[arg_index], "B9600"))
					baudrate = B9600;
				else if (0 == strcmp(argv[arg_index], "B4800"))
					baudrate = B4800;
				else if (0 == strcmp(argv[arg_index], "B2400"))
					baudrate = B2400;
				else if (0 == strcmp(argv[arg_index], "B1800"))
					baudrate = B1800;
				else if (0 == strcmp(argv[arg_index], "B1200"))
					baudrate = B1200;
				else if (0 == strcmp(argv[arg_index], "B600"))
					baudrate = B600;
				else if (0 == strcmp(argv[arg_index], "B300"))
					baudrate = B300;
				else if (0 == strcmp(argv[arg_index], "B150"))
					baudrate = B150;
				else
				{
					snprintf(tmpText, sizeof(tmpText), "wrong parameter: %s, try -? for help", argv[arg_index]);
					return print_error(tmpText);
				}
			}
			else
				return print_error("missing parameter, try -? for help");
		}
		else if (0 == strcmp(argv[arg_index], "-d"))
		{
			arg_index++;
			if (arg_index < argc)
				strncpy(device, argv[arg_index], sizeof(device) - 1);
			else
				return print_error("missing parameter, try -? for help");
		}
		else if (0 == strcmp(argv[arg_index], "-c"))
		{
			if (command_length == 0)
			{
				arg_index++;
				if (arg_index < argc)
					strncpy(command, argv[arg_index], sizeof(command) - 1);
				strcat(command, "\r\n");
				command_length = strlen(command);
			}
			else
				return print_error("wrong parameter, try -? for help");
		}
		else if (0 == strcmp(argv[arg_index], "-x"))
		{
			if (command_length == 0)
			{
				arg_index++;
				if (arg_index < argc)
				{
					char buf[100];
					char *p;
					int i;

					for (i = 0, p = argv[arg_index]; i < (sizeof(buf) - 1) && *p; p++)
					{
						if (isxdigit(*p))
							buf[i++] = *p;
					}
					buf[i] = 0;

					for (p = buf; i >= 2; p += 2, i -= 2)
					{
						int ch;
						if (sscanf(p, "%2x", &ch) == 1)
							command[command_length++] = ch;
					}
				}
				break; //cancel current loop
			}
			else
				return print_error("wrong parameter, try -? for help");
		}
		else if (0 == strcmp(argv[arg_index], "-t"))
		{
			arg_index++;
			if (arg_index < argc)
#ifdef TIMESEC
				timeout = strtoul(argv[arg_index], NULL, 0);
#else
				timeout = 1000 * strtoul(argv[arg_index], NULL, 0);
#endif
			else
				return print_error("missing parameter, try -? for help");
		}
#ifndef TIMESEC
		else if (0 == strcmp(argv[arg_index], "-T"))
		{
			arg_index++;
			if (arg_index < argc)
				timeout = strtoul(argv[arg_index], NULL, 0);
			else
				return print_error("missing parameter, try -? for help");
		}
#endif
		else if (0 == strcmp(argv[arg_index], "-s"))
		{
			arg_index++;
			if (arg_index < argc)
				command_delay = strtoul(argv[arg_index], NULL, 0);
			else
				return print_error("missing parameter, try -? for help");
		}
		else if (0 == strcmp(argv[arg_index], "-m"))
			enable_rs485 = true;
		else if (0 == strcmp(argv[arg_index], "-n"))
			enable_rs485_echo = true;
		else if (0 == strcmp(argv[arg_index], "-e"))
			enable_7E1 = true;
		else if (0 == strcmp(argv[arg_index], "-f"))
			enable_8E1 = true;
		else if (0 == strcmp(argv[arg_index], "-l"))
			display_modem = true;
		else if (0 == strcmp(argv[arg_index], "-p"))
			plc_with_t_req = true;
		else if (0 == strcmp(argv[arg_index], "-r"))
			display_raw = true;
		else if (0 == strcmp(argv[arg_index], "-o"))
			enable_rtscts = true;
		else if (0 == strcmp(argv[arg_index], "-h"))
			return print_help();
		else if (0 == strcmp(argv[arg_index], "-?"))
			return print_help();
		else
		{
			snprintf(tmpText, sizeof(tmpText), "wrong parameter: %s, try -? for help", argv[arg_index]);
			return print_error(tmpText);
		}
		arg_index++;
	}

	if (strlen(device))
	{
		T_TTY_IO tty;
		io_init(&tty);

		fd = open(device, O_RDWR | O_NOCTTY);
		if (fd > 0)
		{
			int rc, flags;

			if (baudrate >= B460800)
				baudrate = B460800;
			else if (baudrate >= B230400)
				baudrate = B230400;
			else if (baudrate >= B115200)
				baudrate = B115200;
			else if (baudrate >= B57600)
				baudrate = B57600;
			else if (baudrate >= B38400)
				baudrate = B38400;
			else if (baudrate >= B19200)
				baudrate = B19200;
			else if (baudrate >= B9600)
				baudrate = B9600;
			else if (baudrate >= B4800)
				baudrate = B4800;
			else if (baudrate >= B2400)
				baudrate = B2400;
			else if (baudrate >= B1800)
				baudrate = B1800;
			else if (baudrate >= B1200)
				baudrate = B1200;
			else if (baudrate >= B600)
				baudrate = B600;
			else if (baudrate >= B300)
				baudrate = B300;
			else
				baudrate = B150;

			rc = tty_init(fd, baudrate, enable_rtscts, enable_7E1, enable_8E1);
			if (rc < 0)
			{
				snprintf(tmpText, sizeof(tmpText), "can't init %s", device);
				return print_error(tmpText);
			}

			if (enable_rs485)
				tty_set_RS485(fd, baudrate, true, enable_rs485_echo);

			flags = fcntl(fd, F_GETFL);
			if (flags < 0)
				return print_error("can't enable nonblocking");
			flags |= O_NONBLOCK; //enable nonblocking
			flags = fcntl(fd, F_SETFL, flags);
			if (flags < 0)
				return print_error("can't enable nonblocking");

			if (display_modem)
				get_tty_modem_lines(fd);

			if (plc_with_t_req)
			{
				tty_set_RTS(fd, true);
				usleep(100);
				tty_set_RTS(fd, false);
			}

			if (command_length > 0)
			{
				if (command_delay > 0)
				{
					//wait some time to power up the mbus device
					usleep(command_delay*1000);
				}
				rc = write(fd, command, command_length);
				if (rc != command_length)
					return print_error("can't write");
			}

			while (demo_running)
			{
#ifdef TIMESEC
				if (timeout < difftime(time((time_t *) 0), start_time))
					break;
#else
				gettimeofday(&t2, NULL);
				int milliSeconds = (t2.tv_sec - t1.tv_sec) * 1000 + (t2.tv_usec - t1.tv_usec)/1000;
				if (timeout < milliSeconds)
					break;
#endif

				if (demo_select(fd, &tty, display_raw) < 0)
					break;
			}
			tty_set_DTR(fd, 0);
			close(fd);

			if (0 < (rc = io_get_data(&tty, tmpText, sizeof(tmpText))))
			{
				tmpText[rc] = 0;
				printf("received: %s\n", tmpText);
			}
		}
		else
		{
			snprintf(tmpText, sizeof(tmpText), "can't open %s", device);
			return print_error(tmpText);
		}
	}
	else
		return print_error("missing device name, try -? for help");

#ifdef MEMLIB_ENABLED
	memlib_showorphans();
#endif
	return EXIT_SUCCESS;
}
